"""Toolbox for the ``ee.ImageCollection`` class."""
from __future__ import annotations

import uuid
from datetime import datetime as dt

import ee
import ee_extra
import requests
import xarray
from ee import apifunction
from matplotlib.axes import Axes
from xarray import Dataset
from xee.ext import REQUEST_BYTE_LIMIT

from .accessors import register_class_accessor
from .utils import plot_data

PY_DATE_FORMAT = "%Y-%m-%dT%H-%M-%S"
"The python format to use to parse dates coming from GEE."

EE_DATE_FORMAT = "YYYY-MM-dd'T'HH-mm-ss"
"The javascript format to use to burn date object in GEE."


@register_class_accessor(ee.ImageCollection, "geetools")
class ImageCollectionAccessor:
    """Toolbox for the ``ee.ImageCollection`` class."""

    def __init__(self, obj: ee.ImageCollection):
        """Instantiate the class."""
        self._obj = obj

    # -- ee-extra wrapper ------------------------------------------------------
    def maskClouds(
        self,
        method: str = "cloud_prob",
        prob: int = 60,
        maskCirrus: bool = True,
        maskShadows: bool = True,
        scaledImage: bool = False,
        dark: float = 0.15,
        cloudDist: int = 1000,
        buffer: int = 250,
        cdi: int | None = None,
    ) -> ee.ImageCollection:
        """Masks clouds and shadows in each image of an ImageCollection (valid just for Surface Reflectance products).

        Parameters:
            self: ImageCollection to mask.
            method: Method used to mask clouds. This parameter is ignored for Landsat products.
                Available options:
                    - 'cloud_prob' : Use cloud probability.
                    - 'qa' : Use Quality Assessment band.
            prob: Cloud probability threshold. Valid just for method = 'cloud_prob'. This parameter is ignored for Landsat products.
            maskCirrus: Whether to mask cirrus clouds. Default to ``True``. Valid just for method = 'qa'. This parameter is ignored for Landsat products.
            maskShadows: Whether to mask cloud shadows. Default to ``True`` This parameter is ignored for Landsat products.
            scaledImage: Whether the pixel values are scaled to the range [0,1] (reflectance values). This parameter is ignored for Landsat products.
            dark: NIR threshold. NIR values below this threshold are potential cloud shadows. This parameter is ignored for Landsat products.
            cloudDist: Maximum distance in meters (m) to look for cloud shadows from cloud edges. This parameter is ignored for Landsat products.
            buffer: Distance in meters (m) to dilate cloud and cloud shadows objects. This parameter is ignored for Landsat products.
            cdi: Cloud Displacement Index threshold. Values below this threshold are considered potential clouds. A cdi = None means that the index is not used. This parameter is ignored for Landsat products.

        Returns:
            Cloud-shadow masked image.

        Notes:
            This method may mask water as well as clouds for the Sentinel-3 Radiance product.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()
                S2 = (
                    ee.ImageCollection('COPERNICUS/S2_SR')
                    .maskClouds(prob = 75,buffer = 300,cdi = -0.5)
                    .first()
                )

        """
        return ee_extra.QA.clouds.maskClouds(
            self._obj,
            method,
            prob,
            maskCirrus,
            maskShadows,
            scaledImage,
            dark,
            cloudDist,
            buffer,
            cdi,
        )

    def closest(
        self, date: ee.Date | str, tolerance: int = 1, unit: str = "month"
    ) -> ee.ImageCollection:
        """Gets the closest image (or set of images if the collection intersects a region that requires multiple scenes) to the specified date.

        Parameters:
            date: Date of interest. The method will look for images closest to this date.
            tolerance: Filter the collection to [date - tolerance, date + tolerance) before searching the closest image. This speeds up the searching process for collections with a high temporal resolution.
            unit: Units for tolerance. Available units: 'year', 'month', 'week', 'day', 'hour', 'minute' or 'second'.

        Returns:
            Closest images to the specified date.

        Examples:
            .. code-block:: python

                import ee
                import geetools

                s2 = ee.ImageCollection('COPERNICUS/S2_SR').closest('2020-10-15')
                s2.size().getInfo()
        """
        return ee_extra.ImageCollection.core.closest(self._obj, date, tolerance, unit)

    def spectralIndices(
        self,
        index: str = "NDVI",
        G: float | int = 2.5,
        C1: float | int = 6.0,
        C2: float | int = 7.5,
        L: float | int = 1.0,
        cexp: float | int = 1.16,
        nexp: float | int = 2.0,
        alpha: float | int = 0.1,
        slope: float | int = 1.0,
        intercept: float | int = 0.0,
        gamma: float | int = 1.0,
        omega: float | int = 2.0,
        beta: float | int = 0.05,
        k: float | int = 0.0,
        fdelta: float | int = 0.581,
        kernel: str = "RBF",
        sigma: str = "0.5 * (a + b)",
        p: float | int = 2.0,
        c: float | int = 1.0,
        lambdaN: float | int = 858.5,
        lambdaR: float | int = 645.0,
        lambdaG: float | int = 555.0,
        online: bool = False,
    ) -> ee.ImageCollection:
        """Computes one or more spectral indices (indices are added as bands) for an image from the Awesome List of Spectral Indices.

        Parameters:
            self: Image to compute indices on. Must be scaled to [0,1].
            index: Index or list of indices to compute, default = 'NDVI'
                Available options:
                    - 'vegetation' : Compute all vegetation indices.
                    - 'burn' : Compute all burn indices.
                    - 'water' : Compute all water indices.
                    - 'snow' : Compute all snow indices.
                    - 'urban' : Compute all urban (built-up) indices.
                    - 'kernel' : Compute all kernel indices.
                    - 'all' : Compute all indices listed below.
                    - Awesome Spectral Indices for GEE: Check the complete list of indices `here <https://awesome-ee-spectral-indices.readthedocs.io/en/latest/list.html>`_.
            G: Gain factor. Used just for index = 'EVI', default = 2.5
            C1: Coefficient 1 for the aerosol resistance term. Used just for index = 'EVI', default = 6.0
            C2: Coefficient 2 for the aerosol resistance term. Used just for index = 'EVI', default = 7.5
            L: Canopy background adjustment. Used just for index = ['EVI','SAVI'], default = 1.0
            cexp: Exponent used for OCVI, default = 1.16
            nexp: Exponent used for GDVI, default = 2.0
            alpha: Weighting coefficient used for WDRVI, default = 0.1
            slope: Soil line slope, default = 1.0
            intercept: Soil line intercept, default = 0.0
            gamma: Weighting coefficient used for ARVI, default = 1.0
            omega: Weighting coefficient  used for MBWI, default = 2.0
            beta: Calibration parameter used for NDSIns, default = 0.05
            k: Slope parameter by soil used for NIRvH2, default = 0.0
            fdelta: Adjustment factor used for SEVI, default = 0.581
            kernel: Kernel used for kernel indices, default = 'RBF'
                Available options:
                    - 'linear' : Linear Kernel.
                    - 'RBF' : Radial Basis Function (RBF) Kernel.
                    - 'poly' : Polynomial Kernel.
            sigma: Length-scale parameter. Used for kernel = 'RBF', default = '0.5 * (a + b)'. If str, this must be an expression including 'a' and 'b'. If numeric, this must be positive.
            p: Kernel degree. Used for kernel = 'poly', default = 2.0
            c: Free parameter that trades off the influence of higher-order versus lower-order terms in the polynomial kernel. Used for kernel = 'poly', default = 1.0. This must be greater than or equal to 0.
            lambdaN: NIR wavelength used for NIRvH2 and NDGI, default = 858.5
            lambdaR: Red wavelength used for NIRvH2 and NDGI, default = 645.0
            lambdaG: Green wavelength used for NDGI, default = 555.0
            drop: Whether to drop all bands except the new spectral indices, default = False

        Returns:
            Image with the computed spectral index, or indices, as new bands.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()
                image = ee.Image('COPERNICUS/S2_SR/20190828T151811_20190828T151809_T18GYT')
                image = image.geetools.specralIndices(["NDVI", "NDFI"])
        """
        # fmt: off
        return ee_extra.Spectral.core.spectralIndices(
            self._obj, index, G, C1, C2, L, cexp, nexp, alpha, slope, intercept, gamma, omega,
            beta, k, fdelta, kernel, sigma, p, c, lambdaN, lambdaR, lambdaG, online,
            drop=False,
        )
        # fmt: on

    def getScaleParams(self) -> dict:
        """Gets the scale parameters for each band of the image.

        Returns:
            Dictionary with the scale parameters for each band.


        Examples:
            .. code-block:: python

                import ee
                import geetools

                ee.Initialize()

                ee.ImageCollection('MODIS/006/MOD11A2').geetools.getScaleParams()
        """
        return ee_extra.STAC.core.getScaleParams(self._obj)

    def getOffsetParams(self) -> dict:
        """Gets the offset parameters for each band of the image.

        Returns:
            Dictionary with the offset parameters for each band.

        Examples:
            .. code-block:: python

            import ee
            import geetools

            ee.Initialize()

            ee.ImageCollection('MODIS/006/MOD11A2').getOffsetParams()
        """
        return ee_extra.STAC.core.getOffsetParams(self._obj)

    def scaleAndOffset(self) -> ee.ImageCollection:
        """Scales bands on an image according to their scale and offset parameters.

        Returns:
            Scaled image.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                S2 = ee.ImageCollection('COPERNICUS/S2_SR').scaleAndOffset()
        """
        return ee_extra.STAC.core.scaleAndOffset(self._obj)

    def preprocess(self, **kwargs) -> ee.ImageCollection:
        """Pre-processes the image: masks clouds and shadows, and scales and offsets the image.

        Parameters:
            **kwargs: Keywords arguments for ``maskClouds`` method.

        Returns:
            Pre-processed image.

        Examples:
            .. code-block:: python

            import ee
            import geetools

            ee.Initialize()
            S2 = ee.ImageCollection('COPERNICUS/S2_SR').preprocess()
        """
        return ee_extra.QA.pipelines.preprocess(self._obj, **kwargs)

    def getSTAC(self) -> dict:
        """Gets the STAC of the image.

        Returns:
            STAC of the image.

        Examples:
            .. code-block:: python

            import ee
            import geetools

            ee.Initialize()

            ee.ImageCollection('COPERNICUS/S2_SR').getSTAC()
        """
        # extract the Asset id from the imagecollection
        assetId = self._obj.get("system:id").getInfo()

        # search for the project in the GEE catalog and extract the project catalog URL
        project = assetId.split("/")[0]
        catalog = "https://earthengine-stac.storage.googleapis.com/catalog/catalog.json"
        links = requests.get(catalog).json()["links"]
        project_catalog = next((i["href"] for i in links if i.get("title") == project), None)
        if project_catalog is None:
            raise ValueError(f"Project {project} not found in the catalog")

        # search for the collection in the project catalog and extract the collection STAC URL
        collection = "_".join(assetId.split("/"))
        links = requests.get(project_catalog).json()["links"]
        collection_stac = next((i["href"] for i in links if i.get("title") == collection), None)
        if collection_stac is None:
            raise ValueError(f"Collection {collection} not found in the {project} catalog")

        return requests.get(collection_stac).json()

    def getDOI(self) -> str:
        """Gets the DOI of the collection, if available.

        Returns:
            DOI of the ee.Image dataset.

        Examples:
            .. code-block:: python

                import ee
                import geetools

                ee.Initialize()

                ee.ImageCollection('NASA/GPM_L3/IMERG_V06').getDOI()
        """
        return ee_extra.STAC.core.getDOI(self._obj)

    def getCitation(self) -> str:
        """Gets the citation of the image, if available.

        Returns:
            Citation of the ee.Image dataset.

        Examples:
            .. code-block:: python

                import ee
                import geetools

                ee.Initialize()

                ee.ImageCollection('NASA/GPM_L3/IMERG_V06').getCitation()
        """
        return ee_extra.STAC.core.getCitation(self._obj)

    def panSharpen(self, method: str = "SFIM", qa: str = "", **kwargs) -> ee.ImageCollection:
        """Apply panchromatic sharpening to the ImageCollection images.

        Optionally, run quality assessments between the original and sharpened Image to
        measure spectral distortion and set results as properties of the sharpened Image.

        Parameters:
            method: The sharpening algorithm to apply. Current options are "SFIM" (Smoothing Filter-based Intensity Modulation), "HPFA" (High Pass Filter Addition), "PCS" (Principal Component Substitution), and "SM" (simple mean). Different sharpening methods will produce different quality sharpening results in different scenarios.
            qa: One or more optional quality assessment names to apply after sharpening. Results will be stored as image properties with the pattern `geetools:metric`, e.g. `geetools:RMSE`.
            **kwargs: Keyword arguments passed to ee.Image.reduceRegion() such as "geometry", "maxPixels", "bestEffort", etc. These arguments are only used for PCS sharpening and quality assessments.

        Returns:
            The ImageCollections with all sharpenable bands sharpened to the panchromatic resolution and quality assessments run and set as properties.

        Examples:
            .. code-block:: python

                import ee
                import geetools

                ee.Initialize()

                source = ee.Image("LANDSAT/LC08/C01/T1_TOA/LC08_047027_20160819")
                sharp = source.panSharpen(method="HPFA", qa=["MSE", "RMSE"], maxPixels=1e13)
        """
        return ee_extra.Algorithms.core.panSharpen(
            img=self._obj, method=method, qa=qa or None, prefix="geetools", **kwargs
        )

    def tasseledCap(self) -> ee.ImageCollection:
        """Calculates tasseled cap brightness, wetness, and greenness components.

        Tasseled cap transformations are applied using coefficients published for these
        supported platforms:

        * Sentinel-2 MSI Level 1C
        * Landsat 9 OLI-2 SR
        * Landsat 9 OLI-2 TOA
        * Landsat 8 OLI SR
        * Landsat 8 OLI TOA
        * Landsat 7 ETM+ TOA
        * Landsat 5 TM Raw DN
        * Landsat 4 TM Raw DN
        * Landsat 4 TM Surface Reflectance
        * MODIS NBAR

        Parameters:
            self: ee.ImageCollection to calculate tasseled cap components for. Must belong to a supported platform.

        Returns:
            ImageCollections with the tasseled cap components as new bands.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                image = ee.Image('COPERNICUS/S2_SR')
                img = img.tasseledCap()
        """
        return ee_extra.Spectral.core.tasseledCap(self._obj)

    def append(self, image: ee.Image) -> ee.ImageCollection:
        """Append an image to the existing image collection.

        Args:
            image: Image to append to the collection.

        Returns:
            ImageCollection with the new image appended.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                ic = ee.ImageCollection('COPERNICUS/S2_SR');

                geom = ee.Geometry.Point(-122.196, 41.411);
                ic2018 = ic.filterBounds(geom).filterDate('2019-07-01', '2019-10-01')
                ic2021 = ic.filterBounds(geom).filterDate('2021-07-01', '2021-10-01')

                ic = ic2018.append(ic2021.first())
                ic.getInfo()
        """
        return self._obj.merge(ee.ImageCollection([image]))

    def collectionMask(self) -> ee.Image:
        """A binary ee.Image where only pixels that are masked in all images of the collection get masked.

        Returns:
            ee.Image of the mask. 1 where at least 1 pixel is valid 0 elswere

        Examples:
            .. code-block::

                import ee, geetools

                ee.Initialize()

                ic = ee.ImageCollection('COPERNICUS/S2_SR');

                geom = ee.Geometry.Point(-122.196, 41.411);
                ic2018 = ic.filterBounds(geom).filterDate('2019-07-01', '2019-10-01')
                ic = ic2018.geetools.collectionMask()
                ic.getInfo()
        """
        masks = self._obj.map(lambda i: i.mask())
        return ee.Image(masks.sum().gt(0))

    def iloc(self, index: int) -> ee.Image:
        """Get Image from the ImageCollection by index.

        Args:
            index: Index of the image to get.

        Returns:
            ee.Image at the specified index.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                ic = ee.ImageCollection('COPERNICUS/S2_SR');

                geom = ee.Geometry.Point(-122.196, 41.411);
                ic2018 = ic.filterBounds(geom).filterDate('2019-07-01', '2019-10-01')
                ic2018.geetools.iloc(0).getInfo()
        """
        return ee.Image(self._obj.toList(self._obj.size()).get(index))

    def integral(self, band: str, time: str = "system:time_start", unit: str = "") -> ee.Image:
        """Compute the integral of a band over time or a specified property.

        Args:
            band: the name of the band to integrate
            time: the name of the property to use as time. It must be a date property of the images.
            unit: the time unit use to compute the integral. It can be one of the following: ["year", "month", "day", "hour", "minute", "second"]. If non is set, the time will be normalized on the integral length.

        Returns:
            An Image object with the integrated band for each pixel

        Examples:
            .. code-block:: python

                import ee, LDCGEETools

                collection = (
                    ee.ImageCollection("LANDSAT/LC08/C01/T1_TOA")
                    .filterBounds(ee.Geometry.Point(-122.262, 37.8719))
                    .filterDate("2014-01-01", "2014-12-31")
                )

                integral = collection.geetools.integral("B1")
                print(integral.getInfo())
        """
        # compute the intervals along the x axis
        # the GEE time is stored as a milliseconds timestamp. If the time unit is not set,
        # the integral is normalized on the total time length of the time series
        minTime = self._obj.aggregate_min(time)
        maxTime = self._obj.aggregate_max(time)
        intervals = {
            "year": ee.Number(1000 * 60 * 60 * 24 * 365),  # 1 year in milliseconds
            "month": ee.Number(1000 * 60 * 60 * 24 * 30),  # 1 month in milliseconds
            "day": ee.Number(1000 * 60 * 60 * 24),  # 1 day in milliseconds
            "hour": ee.Number(1000 * 60 * 60),  # 1 hour in milliseconds
            "minute": ee.Number(1000 * 60),  # 1 minute in milliseconds
            "second": ee.Number(1000),  # 1 second in milliseconds
            "": ee.Number(maxTime).subtract(ee.Number(minTime)),
        }
        interval = intervals[unit]

        # initialize the sum with a 0 value initial item
        # all the properties of the first image of the collection are copied
        first = self._obj.first()
        zero = ee.Image.constant(0).copyProperties(first, first.propertyNames())
        s = ee.Image(zero).rename("integral").set("last", zero)

        # compute the approximation of the integral using the trapezoidal method
        # each local interval is aproximated by the corresponding trapez and the
        # sum is updated
        def computeIntegral(image, integral):
            image = ee.Image(image).select(band)
            integral = ee.Image(integral)
            last = ee.Image(integral.get("last"))
            locMinTime = ee.Number(last.get(time))
            locMaxTime = ee.Number(image.get(time))
            locInterval = locMaxTime.subtract(locMinTime).divide(interval)
            locIntegral = last.add(image).multiply(locInterval).divide(2)
            return integral.add(locIntegral).set("last", image)

        return ee.Image(self._obj.iterate(computeIntegral, s))

    def outliers(
        self, bands: list | ee.List = [], sigma: float | int | ee.Number = 2, drop: bool = False
    ) -> ee.ImageCollection:
        """Compute the outlier for each pixel in the specified bands.

        A pixel is considered as an outlier if:

        .. code-block::

            outlier = value > mean+(sigma*stddev)
            outlier = value < mean-(sigma*stddev)

        In a 1D example it would be:
        - values = [1, 5, 6, 4, 7, 10]
        - mean = 5.5
        - std dev = 3
        - mean + (sigma*stddev) = 8.5
        - mean - (sigma*stddev) = 2.5
        - outliers = values between 2.5 and 8.5 = [1, 10]

        Here in this function an extra band is added to each image for each of the evaluated bands with the outlier status. The band name is the original band name with the suffix "_outlier". A value of 1 means that the pixel is an outlier, 0 means that it is not.

        Optionally users can discard this band by setting ``drop`` to ``True`` and the outlier will simply be masked from each ilmage. This is useful when the outlier band is not needed and the user wants to save space.

        idea from: https://www.kdnuggets.com/2017/02/removing-outliers-standard-deviation-python.html

        Args:
            bands: the bands to evaluate for outliers. If empty, all bands are evaluated
            sigma: the number of standard deviations to use to compute the outlier
            drop: whether to drop the outlier band from the images

        Returns:
            an ImageCollection with the outlier band added to each image or masked if ``drop`` is ``True``

        Examples:
            .. code-block:: python

                import ee, LDCGEETools

                collection = (
                    ee.ImageCollection("LANDSAT/LC08/C01/T1_TOA")
                    .filterBounds(ee.Geometry.Point(-122.262, 37.8719))
                    .filterDate("2014-01-01", "2014-12-31")
                )

                outliers = collection.geetools.outliers(["B1", "B2"], 2)
                print(outliers.getInfo())
        """
        # cast parameters and compute the outlier band names
        initBands = self._obj.first().bandNames()
        statBands = ee.List(bands) if bands else initBands
        outBands = statBands.map(lambda b: ee.String(b).cat("_outlier"))

        # compute the mean and std dev for each band
        statCollection = self._obj.select(statBands)
        mean = statCollection.mean()
        stdDev = statCollection.reduce(ee.Reducer.stdDev())
        minValues = mean.subtract(stdDev.multiply(sigma))
        maxValues = mean.add(stdDev.multiply(sigma))

        # compute the outlier band for each image
        def computeOutlierBands(i):
            outImage = i.select(statBands)
            outImage = outImage.gt(maxValues).Or(outImage.lt(minValues))
            return i.addBands(outImage.rename(outBands))

        ic = self._obj.map(computeOutlierBands)

        # drop the outlier band and mask each image if requested
        def maskOutliers(i):
            maskedBands = i.select(statBands).updateMask(i.select(outBands).Not())
            return i.addBands(maskedBands, overwrite=True).select(initBands)

        ic = ic if drop is False else ic.map(maskOutliers)

        return ee.ImageCollection(ic)

    def to_xarray(
        self,
        drop_variables: tuple[str, ...] | None = None,
        io_chunks: object = None,
        n_images: int = -1,
        mask_and_scale: bool = True,
        decode_times: bool = True,
        decode_timedelta: bool | None = None,
        use_cftime: bool | None = None,
        concat_characters: bool = True,
        decode_coords: bool = True,
        crs: str | None = None,
        scale: float | int | None = None,
        projection: ee.Projection | None = None,
        geometry: ee.Geometry | None = None,
        primary_dim_name: str | None = None,
        primary_dim_property: str | None = None,
        ee_mask_value: float | None = None,
        request_byte_limit: int = REQUEST_BYTE_LIMIT,
    ) -> Dataset:
        """Open an Earth Engine ImageCollection as an ``xarray.Dataset``.

        Args:
            drop_variables: Variables or bands to drop before opening.
            io_chunks: Specifies the chunking strategy for loading data from EE. By default, this automatically calculates optional chunks based on the ``request_byte_limit``.
            n_images: The max number of EE images in the collection to open. Useful when there are a large number of images in the collection since calculating collection size can be slow. -1 indicates that all images should be included.
            mask_and_scale: Lazily scale (using scale_factor and add_offset) and mask (using _FillValue).
            decode_times: Decode cf times (e.g., integers since "hours since 2000-01-01") to np.datetime64.
            decode_timedelta: If True, decode variables and coordinates with time units in {"days", "hours", "minutes", "seconds", "milliseconds", "microseconds"} into timedelta objects. If False, leave them encoded as numbers. If None (default), assume the same value of decode_time.
            use_cftime: Only relevant if encoded dates come from a standard calendar (e.g. "gregorian", "proleptic_gregorian", "standard", or not specified).  If None (default), attempt to decode times to ``np.datetime64[ns]`` objects; if this is not possible, decode times to ``cftime.datetime`` objects. If True, always decode times to ``cftime.datetime`` objects, regardless of whether or not they can be represented using ``np.datetime64[ns]`` objects.  If False, always decode times to ``np.datetime64[ns]`` objects; if this is not possible raise an error.
            concat_characters: Should character arrays be concatenated to strings, for example: ["h", "e", "l", "l", "o"] -> "hello"
            decode_coords: bool or {"coordinates", "all"}, Controls which variables are set as coordinate variables: - "coordinates" or True: Set variables referred to in the ``'coordinates'`` attribute of the datasets or individual variables as coordinate variables. - "all": Set variables referred to in  ``'grid_mapping'``, ``'bounds'`` and other attributes as coordinate variables.
            crs: The coordinate reference system (a CRS code or WKT string). This defines the frame of reference to coalesce all variables upon opening. By default, data is opened with 'EPSG:4326'.
            scale: The scale in the ``crs`` or ``projection``'s units of measure -- either meters or degrees. This defines the scale that all data is represented in upon opening. By default, the scale is 1° when the CRS is in degrees or 10,000 when in meters.
            projection: Specify an ``ee.Projection`` object to define the ``scale`` and ``crs`` (or other coordinate reference system) with which to coalesce all variables upon opening. By default, the scale and reference system is set by the the ``crs`` and ``scale`` arguments.
            geometry: Specify an ``ee.Geometry`` to define the regional bounds when opening the data. When not set, the bounds are defined by the CRS's ``area_of_use`` boundaries. If those aren't present, the bounds are derived from the geometry of the first image of the collection.
            primary_dim_name: Override the name of the primary dimension of the output Dataset. By default, the name is 'time'.
            primary_dim_property: Override the ``ee.Image`` property for which to derive the values of the primary dimension. By default, this is 'system:time_start'.
            ee_mask_value: Value to mask to EE nodata values. By default, this is 'np.iinfo(np.int32).max' i.e. 2147483647.
            request_byte_limit: the max allowed bytes to request at a time from Earth Engine. By default, it is 48MBs.

        Returns:
            An ``xarray.Dataset`` that streams in remote data from Earth Engine.
        """
        return xarray.open_dataset(
            self._obj,
            engine="ee",
            drop_variables=drop_variables,
            io_chunks=io_chunks,
            n_images=n_images,
            mask_and_scale=mask_and_scale,
            decode_times=decode_times,
            decode_timedelta=decode_timedelta,
            use_cftime=use_cftime,
            concat_characters=concat_characters,
            decode_coords=decode_coords,
            crs=crs,
            scale=scale,
            projection=projection,
            geometry=geometry,
            primary_dim_name=primary_dim_name,
            primary_dim_property=primary_dim_property,
            ee_mask_value=ee_mask_value,
            request_byte_limit=request_byte_limit,
        )

    def validPixel(self, band: str | ee.String = "") -> ee.Image:
        """Compute the number of valid pixels in the specified band.

        Compute the number of valid pixels in the specified band. 2 bands will be created:
        one with the number of valid pixels (``valid``) and another with the percentage of valid pixels (``pct_valid``).

        Args:
            band: the band to evaluate for valid pixels. If empty, use the first band
        Returns:
            an Image with the number of valid pixels or the percentage of valid pixels.

        Examples:
            .. code-block:: python
                import ee, LDCGEETools
                collection = (
                    ee.ImageCollection("LANDSAT/LC08/C01/T1_TOA")
                    .filterBounds(ee.Geometry.Point(-122.262, 37.8719))
                    .filterDate("2014-01-01", "2014-12-31")
                )
                valid = collection.geetools.validPixels("B1")
                print(valid.getInfo()).
        """
        # compute the mask for the specified band
        band = self._obj.first().bandNames().get(0) if band == "" else ee.String(band)
        masks = self._obj.select([band]).map(lambda i: i.mask().eq(1))
        validPixel = masks.sum().rename("valid").clip(self._obj.geometry())
        validPct = validPixel.divide(self._obj.size()).multiply(100).rename("pct_valid")
        return validPixel.addBands(validPct)

    def containsBandNames(self, bandNames: list | ee.List, filter: str) -> ee.ImageCollection:
        """Filter the ImageCollection by band names using the provided filter.

        Args:
            bandNames: list of band names to filter
            filter: type of filter to apply. To keep images that contains all the specified bands use "ALL". To get the images including at least one of the specified band use "ANY".

        Returns:
            A filtered ImageCollection

        Examples:
            .. code-block:: python

                import ee, LDCGEETools

                collection = (
                    ee.ImageCollection("LANDSAT/LC08/C01/T1_TOA")
                    .filterBounds(ee.Geometry.Point(-122.262, 37.8719))
                    .filterDate("2014-01-01", "2014-12-31")
                )

                filtered = collection.geetools.containsBandNames(["B1", "B2"], "ALL")
                print(filtered.getInfo())
        """
        # cast parameters
        filter = {"ALL": "Filter.and", "ANY": "Filter.or"}[filter]
        bandNames = ee.List(bandNames)

        # add bands as metadata in a temporary property
        band_name = uuid.uuid4().hex
        ic = self._obj.map(lambda i: i.set(band_name, i.bandNames()))

        # create a filter by combining a listContain filter over all the band names from the
        # user list. Combine them with a "Or" to get a "any" filter and "And" to get a "all".
        # We use a workaround until this is solved: https://issuetracker.google.com/issues/322838709
        filterList = bandNames.map(lambda b: ee.Filter.listContains(band_name, b))
        filterCombination = apifunction.ApiFunction.call_(filter, ee.List(filterList))

        # apply this filter and remove the temporary property. Exclude parameter is additive so
        # we do a blank multiplication to remove all the properties beforhand
        ic = ee.ImageCollection(ic.filter(filterCombination))
        ic = ic.map(lambda i: ee.Image(i.multiply(1).copyProperties(i, exclude=[band_name])))

        return ee.ImageCollection(ic)

    def containsAllBands(self, bandNames: list | ee.List) -> ee.ImageCollection:
        """Filter the ImageCollection keeping only the images with all the provided bands.

        Args:
            bandNames: list of band names to filter

        Returns:
            A filtered ImageCollection

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                collection = (
                    ee.ImageCollection("LANDSAT/LC08/C01/T1_TOA")
                    .filterBounds(ee.Geometry.Point(-122.262, 37.8719))
                    .filterDate("2014-01-01", "2014-12-31")
                )

                filtered = collection.geetools.containsAllBands(["B1", "B2"])
                print(filtered.getInfo())
        """
        return self.containsBandNames(bandNames, "ALL")

    def containsAnyBands(self, bandNames: list | ee.List) -> ee.ImageCollection:
        """Filter the ImageCollection keeping only the images with any of the provided bands.

        Args:
            bandNames: list of band names to filter

        Returns:
            A filtered ImageCollection

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                collection = (
                    ee.ImageCollection("LANDSAT/LC08/C01/T1_TOA")
                    .filterBounds(ee.Geometry.Point(-122.262, 37.8719))
                    .filterDate("2014-01-01", "2014-12-31")
                )

                filtered = collection.geetools.containsAnyBands(["B1", "B2"])
                print(filtered.getInfo())
        """
        return self.containsBandNames(bandNames, "ANY")

    def aggregateArray(self, properties: list | ee.List | None = None) -> ee.Dict:
        """Aggregate the ImageCollection selected properties into a dictionary.

        Args:
            properties: list of properties to aggregate. If None, all properties are aggregated.

        Returns:
            A dictionary with the properties as keys and the aggregated values as values.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                collection = (
                    ee.ImageCollection("LANDSAT/LC08/C01/T1_TOA")
                    .filterBounds(ee.Geometry.Point(-122.262, 37.8719))
                    .filterDate("2014-01-01", "2014-12-31")
                )

                aggregated = collection.geetools.aggregateArray(["CLOUD_COVER", "system:time_start"])
                print(aggregated.getInfo())
        """
        keys = ee.List(properties) if properties is not None else self._obj.first().propertyNames()
        values = keys.map(lambda p: self._obj.aggregate_array(p))
        return ee.Dictionary.fromLists(keys, values)

    def groupInterval(self, unit: str = "month", duration: int = 1) -> ee.List:
        """Transform the ImageCollection into a list of smaller collection of the specified duration.

        For example using unit as "month" and duration as 1, the ImageCollection will be transformed
        into a list of ImageCollection with each ImageCollection containing images for each month.
        Make sure the collection is filtered beforeend to reduce the number of images that needs to be
        processed.

        Args:
            unit: The unit of time to split the collection. Available units: 'year', 'month', 'week', 'day', 'hour', 'minute' or 'second'.
            duration: The duration of each split.

        Returns:
            A list of imagecollection grouped by interval

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                collection = (
                    ee.ImageCollection("LANDSAT/LC08/C01/T1_TOA")
                    .filterBounds(ee.Geometry.Point(-122.262, 37.8719))
                    .filterDate("2014-01-01", "2014-12-31")
                )

                split = collection.geetools.groupInterval("month", 1)
                print(split.getInfo())
        """
        # as everything is relyin on the "system:time_start" property
        # we sort the image collection in the first place. In most collection it will change nothing
        # so free of charge unless for plumbing
        ic = self._obj.sort("system:time_start")

        # transform the interval into a duration in milliseconds
        # I can use the DateRangeAccessor as it's imported earlier in the __init__.py file
        # I don't know if it should be properly imported here, let's see with user feedback
        timeList = ic.aggregate_array("system:time_start")
        start, end = timeList.get(0), timeList.get(-1)
        DateRangeList = ee.DateRange(start, end).geetools.split(duration, unit)
        imageCollectionList = DateRangeList.map(
            lambda dr: ic.filterDate(ee.DateRange(dr).start(), ee.DateRange(dr).end())
        )

        return ee.List(imageCollectionList)

    def reduceInterval(
        self,
        reducer: str | ee.Reducer = "mean",
        unit: str = "month",
        duration: int = 1,
    ) -> ee.ImageCollection:
        """Reduce the images included in the same duration interval using the provided reducer.

        For example using unit as "month" and duration as 1, the ImageCollection will be reduced
        into a new ImageCollection with each image containing the reduced values for each month.
        Make sure the collection is filtered beforehand to reduce the number of images that needs to be
        processed.

        Args:
            reducer: The name of the reducer to use or a Reducer object. Default is "mean".
            unit: The unit of time to split the collection. Available units: 'year', 'month', 'week', 'day', 'hour', 'minute' or 'second'.
            duration: The duration of each split.

        Returns:
            A new ImageCollection with the reduced images.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                collection = (
                    ee.ImageCollection("LANDSAT/LC08/C01/T1_TOA")
                    .filterBounds(ee.Geometry.Point(-122.262, 37.8719))
                    .filterDate("2014-01-01", "2014-12-31")
                )

                reduced = collection.geetools.reduceInterval("mean", "month", 1)
                print(reduced.getInfo())
        """
        # create a list of image collections to be reduced
        # Every subcollection is sorted in case one use the "first" reducer
        imageCollectionList = self.groupInterval(unit, duration)

        # create a reducer from user parameters
        red = getattr(ee.Reducer, reducer)() if isinstance(reducer, str) else reducer

        def reduce(ic):
            timeList = ee.ImageCollection(ic).aggregate_array("system:time_start")
            start, end = timeList.get(0), timeList.get(-1)
            bandNames = ee.ImageCollection(ic).first().bandNames()
            image = ee.ImageCollection(ic).reduce(red).rename(bandNames)
            return image.set("system:time_start", start, "system:time_end", end)

        reducedImagesList = imageCollectionList.map(reduce)

        # set back the original properties
        ic = ee.ImageCollection(reducedImagesList).copyProperties(self._obj)

        return ee.ImageCollection(ic)

    def closestDate(self) -> ee.ImageCollection:
        """Fill masked pixels with the first valid pixel in the stack of images.

        The method will for every image, fill all the pixels with the latest nono masked pixel in the stack of images.
        I requires the image to have a valid "system:time_start" property.
        As the imageCollection will need to be sorted limit the analysis to a reasonable number of image by filtering your data beforehand.

        Returns:
            An ImageCollection with all pixels unmasked in every image.

        Examples:
            .. code:: python

                import ee, geetools

                ee.Initialize()

                collection = (
                    ee.ImageCollection("LANDSAT/LC08/C01/T1_TOA")
                    .filterBounds(ee.Geometry.Point(-122.262, 37.8719))
                    .filterDate("2014-01-01", "2014-12-31")
                )

                filled = collection.geetools.fillWithFirst()
                print(filled.getInfo())
        """
        # retrieve all the time starts as an ordered list to iterate
        timeList = self._obj.aggregate_array("system:time_start").sort()

        # for each time start find all the images thata are before and use the mosaic reducer
        # to only keep the first one with a non masked pixel
        def fill(date):
            return self._obj.filter(ee.Filter.lte("system:time_start", date)).mosaic()

        imageList = timeList.map(fill)

        return ee.ImageCollection(imageList)

    def medoid(self) -> ee.image:
        """Compute the medoid of the ImageCollection.

        The medoid is the image that has the smallest sum of distances to all other images in the collection.
        The distance is computed using the Euclidean distance between the pixels of the images.

        Returns:
            An Image that is the medoid of the ImageCollection.

        Examples:
            .. code:: python

                import ee, geetools

                ee.Initialize()

                collection = (
                    ee.ImageCollection("LANDSAT/LC08/C01/T1_TOA")
                    .filterBounds(ee.Geometry.Point(-122.262, 37.8719))
                    .filterDate("2014-01-01", "2014-12-31")
                )

                medoid = collection.geetools.medoid()
                print(medoid.getInfo())
        """
        # create a random name for the sum of distances band to avoid conflicts
        sumOfDistancesName = uuid.uuid4().hex

        # discover bandname from the first image of the collection
        bandNames = self._obj.first().bandNames()

        # normalize the band used to compute the distance
        # first extract the min and max value of each band pixelwizse along the stac and then
        # normalize the pixel values.
        minMax = self._obj.reduce(ee.Reducer.minMax())

        def normalizeBands(image):
            def normalizeBand(bandName):
                band = image.select([bandName])
                bandMin = minMax.select(ee.String(bandName).cat("_min"))
                bandMax = minMax.select(ee.String(bandName).cat("_max"))
                return band.subtract(bandMin).divide(bandMax.subtract(bandMin))

            return ee.ImageCollection(bandNames.map(normalizeBand)).toBands().rename(bandNames)

        normalized = self._obj.map(normalizeBands)

        # compute the distance between each image and all the others
        def computeSumDistance(image):
            def computeDistance(other):
                return image.subtract(other).pow(2).reduce(ee.Reducer.sum()).sqrt()

            sumDistances = normalized.map(computeDistance).reduce(ee.Reducer.sum())
            return image.addBands(sumDistances.rename(sumOfDistancesName))

        sumDistance = normalized.map(computeSumDistance)

        # use the computed sum of distances as a sorting band for a quality mossaic
        # to get the image with the smallest sum of distances
        medoid = sumDistance.qualityMosaic(sumOfDistancesName)

        return ee.Image(medoid).select(bandNames)

    def datesByBands(
        self,
        region: ee.Geometry,
        reducer: str | ee.Reducer = "mean",
        scale: int = 10000,
        dateProperty: str = "system:time_start",
        bands: list = [],
        labels: list = [],
    ) -> ee.Dictionary:
        """Reduce the data for each image in the collection by bands on a specific region.

        This method is returning a dictionary with all the bands as keys and their reduced value for each date over the specified region as value.

        .. code-block::

            {
                "band1": {"date1": value1, "date2": value2, ...},
                "band2": {"date1": value1, "date2": value2, ...},
                ...
            }

        Parameters:
            region: The region to reduce the data on.
            reducer: The name of the reducer or a reducer object use. Default is "mean".
            scale: The scale in meters to use for the reduction. default is 10000m
            dateProperty: The property to use as date for each image. Default is "system:time_start".
            bands: The bands to reduce. If empty, all bands are reduced.
            labels: The labels to use for the bands. If empty, the bands names are used.

        Returns:
            A dictionary with the reduced values for each band and each date.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                collection = (
                    ee.ImageCollection("LANDSAT/LC08/C02/T1_TOA")
                    .filterBounds(ee.Geometry.Point(-122.262, 37.8719))
                    .filterDate("2014-01-01", "2014-12-31")
                )

                region = ee.Geometry.Point(-122.262, 37.8719).buffer(10000)
                reduced = collection.geetools.datesByBands(region, "mean", 10000, "system:time_start")
                print(reduced.getInfo())
        """
        # cast parameters
        eeBands = ee.List(bands) if len(bands) else self._obj.first().bandNames()
        eeLabels = ee.List(labels) if len(labels) else eeBands

        # recast band names as labels in the source collection
        ic = self._obj.select(eeBands).map(lambda i: i.rename(eeLabels))

        # aggregate all the dates contained in the collection
        dateList = ic.aggregate_array(dateProperty).map(lambda d: ee.Date(d).format(EE_DATE_FORMAT))

        # create a reducer from the specified parameters
        red = getattr(ee.Reducer, reducer)() if isinstance(reducer, str) else reducer

        # create a list of dictionaries with the reduced values for each band
        def reduce(lbl: ee.String) -> ee.Dictionary:
            image = ic.select([lbl]).toBands().rename(dateList)
            return image.reduceRegion(red, region, scale)

        return ee.Dictionary.fromLists(eeLabels, eeLabels.map(reduce))

    def datesByRegions(
        self,
        band: str,
        regions: ee.FeatureCollection,
        label: str = "system:index",
        reducer: str | ee.Reducer = "mean",
        scale: int = 10000,
        dateProperty: str = "system:time_start",
    ) -> ee.Dictionary:
        """Reduce the data for each image in the collection by regions for a single band.

        This method is returning a dictionary with all the regions as keys and their reduced value for each date over the specified region for a specific band as value.

        .. code-block::

            {
                "region1": {"date1": value1, "date2": value2, ...},
                "region2": {"date1": value1, "date2": value2, ...},
                ...
            }

        Parameters:
            band: The band to reduce.
            regions: The regions to reduce the data on.
            label: The property to use as label for each region. Default is "system:index".
            reducer: The name of the reducer or a reducer object use. Default is "mean".
            scale: The scale in meters to use for the reduction. default is 10000m
            dateProperty: The property to use as date for each image. Default is "system:time_start".

        Returns:
            A dictionary with the reduced values for each region and each date.

        Examples:
            .. code-block:: python

            import ee, geetools

            ee.Initialize()

            collection = (
                ee.ImageCollection("LANDSAT/LC08/C01/T1_TOA")
                .filterBounds(ee.Geometry.Point(-122.262, 37.8719))
                .filterDate("2014-01-01", "2014-12-31")
            )

            regions = ee.FeatureCollection([
                ee.Feature(ee.Geometry.Point(-122.262, 37.8719).buffer(10000), {"name": "region1"}),
                ee.Feature(ee.Geometry.Point(-122.262, 37.8719).buffer(20000), {"name": "region2"})
            ])

            reduced = collection.geetools.datesByRegions("B1", regions, "name", "mean", 10000, "system:time_start")
            print(reduced.getInfo())
        """
        # aggregate all the dates of the image collection into bands of a single image
        def to_string(date: ee.Date) -> ee.String:
            return ee.Date(date).format(EE_DATE_FORMAT)

        dateList = self._obj.aggregate_array(dateProperty).map(to_string)

        # reduce the data for each region
        image = self._obj.select([band]).toBands().rename(dateList)
        red = getattr(ee.Reducer, reducer)() if isinstance(reducer, str) else reducer
        reduced = image.reduceRegions(regions, red, scale)

        # create a list of dictionaries for each region and aggregate them into a dictionary
        values = reduced.toList(regions.size()).map(lambda f: ee.Feature(f).toDictionary(dateList))
        keys = ee.List(regions.aggregate_array(label))

        return ee.Dictionary.fromLists(keys, values)

    def doyByBands(
        self,
        region: ee.Geometry,
        spatialReducer: str | ee.Reducer = "mean",
        timeReducer: str | ee.Reducer = "mean",
        scale: int = 10000,
        dateProperty: str = "system:time_start",
        bands: list = [],
        labels: list = [],
    ) -> ee.Dictionary:
        """Aggregate the images that occurs on the same day and then reduce each band on a single region.

        This method is returning a dictionary with all the bands as keys and their reduced value for each day over the specified region as value.

        .. code-block::

            {
                "band1": {"doy1": value1, "doy2": value2, ...},
                "band2": {"doy1": value1, "doy2": value2, ...},
                ...
            }

        Parameters:
            region: The region to reduce the data on.
            spatialReducer: The name of the reducer or a reducer object to use for spatial reduction. Default is "mean".
            timeReducer: The name of the reducer or a reducer object to use for time reduction. Default is "mean".
            scale: The scale in meters to use for the reduction. default is 10000m
            dateProperty: The property to use as date for each image. Default is "system:time_start".
            bands: The bands to reduce. If empty, all bands are reduced.
            labels: The labels to use for the bands. If empty, the bands names are used.

        Returns:
            A dictionary with the reduced values for each band and each day.
        """
        # cast parameters
        bands = ee.List(bands) if len(bands) else self._obj.first().bandNames()
        labels = ee.List(labels) if len(labels) else bands

        # recast band names as labels in the source collection
        ic = self._obj.select(bands).map(lambda i: i.rename(labels))

        # create 2 metadata name as random string to avoid any risk of conflicts
        doy_metadata, size_metadata = uuid.uuid4().hex, uuid.uuid4().hex

        # add the day of year as metadata to each image
        def doy_tag(i: ee.Image) -> ee.Image:
            doy = ee.Date(i.get(dateProperty)).getRelative("day", "year")
            return i.set(doy_metadata, doy)

        ic = self._obj.map(doy_tag)

        # create a list of ImageCollection where every images of the same day are grouped together
        dayList = ee.List.sequence(0, 366)

        def filter_doy(d: ee.Number) -> ee.ImageCollection:
            c = ic.filter(ee.Filter.eq(doy_metadata, d))
            c = c.set(size_metadata, c.size())
            return c.set(doy_metadata, d)

        icList = dayList.map(filter_doy)

        # reduce every sub ImageCollection in the list into images (it's the temporal reduction)
        # and aggregate the result as a single ImageCollection
        timeRed = (
            getattr(ee.Reducer, timeReducer)() if isinstance(timeReducer, str) else timeReducer
        )

        def timeReduce(c: ee.imageCollection) -> ee.image:
            c = ee.ImageCollection(c)
            i = c.reduce(timeRed).rename(labels)
            i = i.set(size_metadata, c.get(size_metadata))
            return i.set(doy_metadata, c.get(doy_metadata))

        ic = ee.ImageCollection(icList.map(timeReduce)).filter(ee.Filter.gt(size_metadata, 0))

        # spatially reduce the generated imagecollection over the region for each band
        doyList = ic.aggregate_array(doy_metadata).map(lambda d: ee.Number(d).int().format())
        spatialRed = (
            getattr(ee.Reducer, spatialReducer)()
            if isinstance(spatialReducer, str)
            else spatialReducer
        )

        def spatialReduce(label: ee.String) -> ee.Dictionary:
            image = ic.select([label]).toBands().rename(doyList)
            return image.reduceRegion(spatialRed, region, scale)

        return ee.Dictionary.fromLists(labels, ee.List(labels).map(spatialReduce))

    def doyByRegions(
        self,
        band: str,
        regions: ee.FeatureCollection,
        label: str = "system:index",
        spatialReducer: str | ee.Reducer = "mean",
        timeReducer: str | ee.Reducer = "mean",
        scale: int = 10000,
        dateProperty: str = "system:time_start",
    ) -> ee.Dictionary:
        """Aggregate the images that occurs on the same day and then reduce a single band on multiple regions.

        This method is returning a dictionary with all the regions as keys and their reduced value for each day over the specified region for a specific band as value.

        .. code-block::

            {
                "region1": {"doy1": value1, "doy2": value2, ...},
                "region2": {"doy1": value1, "doy2": value2, ...},
                ...
            }

        Parameters:
            band: The band to reduce.
            regions: The regions to reduce the data on.
            label: The property to use as label for each region. Default is "system:index".
            spatialReducer: The name of the reducer or a reducer object to use for spatial reduction. Default is "mean".
            timeReducer: The name of the reducer or a reducer object to use for time reduction. Default is "mean".
            scale: The scale in meters to use for the reduction. default is 10000m
            dateProperty: The property to use as date for each image. Default is "system:time_start".

        Returns:
            A dictionary with the reduced values for each region and each day.
        """
        # create 2 metadata name as random string to avoid any risk of conflicts
        doy_metadata, size_metadata = uuid.uuid4().hex, uuid.uuid4().hex

        # add the day of year as metadata to each image
        def doy_tag(i: ee.Image) -> ee.Image:
            doy = ee.Date(i.get(dateProperty)).getRelative("day", "year")
            return i.set(doy_metadata, doy)

        ic = self._obj.select([band]).map(doy_tag)

        # create a list of ImageCollection where every images of the same day are grouped together
        dayList = ee.List.sequence(0, 366)

        def filter_doy(d: ee.Number) -> ee.ImageCollection:
            c = ic.filter(ee.Filter.eq(doy_metadata, d))
            c = c.set(size_metadata, c.size())
            return c.set(doy_metadata, d)

        icList = dayList.map(filter_doy)

        # reduce every sub ImageCollection in the list into images (it's the temporal reduction)
        # and aggregate the result as a single ImageCollection
        timeRed = (
            getattr(ee.Reducer, timeReducer)() if isinstance(timeReducer, str) else timeReducer
        )

        def timeReduce(c: ee.imageCollection) -> ee.image:
            c = ee.ImageCollection(c)
            i = c.reduce(timeRed).rename([band])
            i = i.set(size_metadata, c.get(size_metadata))
            return i.set(doy_metadata, c.get(doy_metadata))

        ic = ee.ImageCollection(icList.map(timeReduce)).filter(ee.Filter.gt(size_metadata, 0))

        # reduce the data for each region
        doyList = ic.aggregate_array(doy_metadata).map(lambda d: ee.Number(d).int().format())
        spatialRed = (
            getattr(ee.Reducer, spatialReducer)()
            if isinstance(spatialReducer, str)
            else spatialReducer
        )
        image = ic.toBands().rename(doyList)
        reduced = image.reduceRegions(regions, spatialRed, scale)

        # create a list of dictionaries for each region and aggregate them into a dictionary
        values = reduced.toList(regions.size()).map(lambda f: ee.Feature(f).toDictionary(doyList))
        keys = ee.List(regions.aggregate_array(label))

        return ee.Dictionary.fromLists(keys, values)

    def doyByYears(
        self,
        band: str,
        region: ee.Geometry,
        reducer: str | ee.Reducer = "mean",
        scale: int = 10000,
        dateProperty: str = "system:time_start",
    ) -> ee.Dictionary:
        """Aggregate for each year on a single region a single band.

        This method is returning a dictionary with all the years as keys and their reduced value for each day over the specified region for a specific band as value.

        .. code-block::

            {
                "year1": {"doy1": value1, "doy2": value2, ...},
                "year2": {"doy1": value1, "doy2": value2, ...},
                ...
            }

        Parameters:
            band: The band to reduce.
            region: The region to reduce the data on.
            reducer: The name of the reducer or a reducer object to use. Default is "mean".
            scale: The scale in meters to use for the reduction. default is 10000m
            dateProperty: The property to use as date for each image. Default is "system:time_start".

        Returns:
            A dictionary with the reduced values for each year and each day.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                collection = (
                    ee.ImageCollection("LANDSAT/LC08/C01/T1_TOA")
                    .filterBounds(ee.Geometry.Point(-122.262, 37.8719))
                    .filterDate("2014-01-01", "2014-12-31")
                )

                reduced = collection.geetools.doyByYears("B1", ee.Geometry.Point(-122.262, 37.8719).buffer(10000), "mean", "mean", 10000, "system:time_start")
                print(reduced.getInfo())
        """
        # add a doy metadata to the images
        doy_metadata, year_metadata = uuid.uuid4().hex, uuid.uuid4().hex

        def date_tag(i: ee.Image) -> ee.Image:
            date = ee.Date(i.get(dateProperty))
            doy = date.getRelative("day", "year")
            year = date.get("year")
            return i.set(doy_metadata, doy).set(year_metadata, year)

        ic = self._obj.select([band]).map(date_tag)

        # create a List of image collection where every images from the same year are grouped together
        yearList = ic.aggregate_array(year_metadata).distinct().sort()
        yearKeys = yearList.map(lambda y: ee.Number(y).int().format())
        red = getattr(ee.Reducer, reducer)() if isinstance(reducer, str) else reducer

        def reduce(year: ee.Number) -> ee.Dictionary:
            c = ic.filter(ee.Filter.eq(year_metadata, year))
            doyList = c.aggregate_array(doy_metadata).map(lambda d: ee.Number(d).int().format())
            return c.toBands().rename(doyList).reduceRegion(red, region, scale)

        return ee.Dictionary.fromLists(yearKeys, yearList.map(reduce))

    def plot_dates_by_bands(
        self,
        region: ee.Geometry,
        reducer: str | ee.Reducer = "mean",
        scale: int = 10000,
        dateProperty: str = "system:time_start",
        bands: list = [],
        labels: list = [],
        colors: list = [],
        ax: Axes | None = None,
    ):
        """Plot the reduced data for each image in the collection by bands on a specific region.

        This method is plotting the reduced data for each image in the collection by bands on a specific region.

        Parameters:
            region: The region to reduce the data on.
            reducer: The name of the reducer or a reducer object to use. Default is "mean".
            scale: The scale in meters to use for the reduction. default is 10000m
            dateProperty: The property to use as date for each image. Default is "system:time_start".
            bands: The bands to reduce. If empty, all bands are reduced.
            labels: The labels to use for the bands. If empty, the bands names are used.
            colors: The colors to use for the bands. If empty, the default colors are used.
            ax: The matplotlib axes to plot the data on. If None, a new figure is created.

        Returns:
            A matplotlib axes with the reduced values for each band and each date.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                collection = (
                    ee.ImageCollection("LANDSAT/LC08/C01/T1_TOA")
                    .filterBounds(ee.Geometry.Point(-122.262, 37.8719))
                    .filterDate("2014-01-01", "2014-12-31")
                )

                region = ee.Geometry.Point(-122.262, 37.8719).buffer(10000)
                collection.geetools.plot_dates_by_bands(region, "mean", 10000, "system:time_start")
        """
        # get the reduced data
        raw_data = self.datesByBands(region, reducer, scale, dateProperty, bands, labels).getInfo()

        # transform all the dates int datetime objects
        def to_date(dict):
            return {dt.strptime(d, PY_DATE_FORMAT): v for d, v in dict.items()}

        data = {l: to_date(dict) for l, dict in raw_data.items()}

        # create the plot
        ax = plot_data("date", data, "Date", colors, ax)

        return ax

    def plot_dates_by_regions(
        self,
        band: str,
        regions: ee.FeatureCollection,
        label: str = "system:index",
        reducer: str | ee.Reducer = "mean",
        scale: int = 10000,
        dateProperty: str = "system:time_start",
        colors: list = [],
        ax: Axes | None = None,
    ):
        """Plot the reduced data for each image in the collection by regions for a single band.

        This method is plotting the reduced data for each image in the collection by regions for a single band.

        Parameters:
            band: The band to reduce.
            regions: The regions to reduce the data on.
            label: The property to use as label for each region. Default is "system:index".
            reducer: The name of the reducer or a reducer object to use. Default is "mean".
            scale: The scale in meters to use for the reduction. default is 10000m
            dateProperty: The property to use as date for each image. Default is "system:time_start".
            colors: The colors to use for the regions. If empty, the default colors are used.
            ax: The matplotlib axes to plot the data on. If None, a new figure is created.

        Returns:
            A matplotlib axes with the reduced values for each region and each date.

        Examples:
            .. code-block:: python

            import ee, geetools

            ee.Initialize()

            collection = (
                ee.ImageCollection("LANDSAT/LC08/C01/T1_TOA")
                .filterBounds(ee.Geometry.Point(-122.262, 37.8719))
                .filterDate("2014-01-01", "2014-12-31")
            )

            regions = ee.FeatureCollection([
                ee.Feature(ee.Geometry.Point(-122.262, 37.8719).buffer(10000), {"name": "region1"}),
                ee.Feature(ee.Geometry.Point(-122.262, 37.8719).buffer(20000), {"name": "region2"})
            ])

            collection.geetools.plot_dates_by_regions("B1", regions, "name", "mean", 10000, "system:time_start")
        """
        # get the reduced data
        raw_data = self.datesByRegions(band, regions, label, reducer, scale, dateProperty).getInfo()

        # transform all the dates int datetime objects
        def to_date(dict):
            return {dt.strptime(d, PY_DATE_FORMAT): v for d, v in dict.items()}

        data = {l: to_date(dict) for l, dict in raw_data.items()}

        # create the plot
        ax = plot_data("date", data, "Date", colors, ax)

        return ax

    def plot_doy_by_bands(
        self,
        region: ee.Geometry,
        spatialReducer: str | ee.Reducer = "mean",
        timeReducer: str | ee.Reducer = "mean",
        scale: int = 10000,
        dateProperty: str = "system:time_start",
        bands: list = [],
        labels: list = [],
        colors: list = [],
        ax: Axes | None = None,
    ):
        """Plot the reduced data for each image in the collection by bands on a specific region.

        This method is plotting the reduced data for each image in the collection by bands on a specific region.

        Parameters:
            region: The region to reduce the data on.
            spatialReducer: The name of the reducer or a reducer object to use. Default is "mean".
            timeReducer: The name of the reducer or a reducer object to use. Default is "mean".
            scale: The scale in meters to use for the reduction. default is 10000m
            dateProperty: The property to use as date for each image. Default is "system:time_start".
            bands: The bands to reduce. If empty, all bands are reduced.
            labels: The labels to use for the bands. If empty, the bands names are used.
            colors: The colors to use for the bands. If empty, the default colors are used.
            ax: The matplotlib axes to plot the data on. If None, a new figure is created.

        Returns:
            A matplotlib axes with the reduced values for each band and each day.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                collection = (
                    ee.ImageCollection("LANDSAT/LC08/C01/T1_TOA")
                    .filterBounds(ee.Geometry.Point(-122.262, 37.8719))
                    .filterDate("2014-01-01", "2014-12-31")
                )

                region = ee.Geometry.Point(-122.262, 37.8719).buffer(10000)
                collection.geetools.plot_doy_by_bands(region, "mean", "mean", 10000, "system:time_start")
        """
        # get the reduced data
        raw_data = self.doyByBands(
            region, spatialReducer, timeReducer, scale, dateProperty, bands, labels
        ).getInfo()

        # transform all the dates strings into int object and reorder the dictionary
        def to_int(d):
            return {int(k): v for k, v in d.items()}

        data = {l: dict(sorted(to_int(raw_data[l]).items())) for l in raw_data}

        # create the plot
        ax = plot_data("doy", data, "Day of Year", colors, ax)

        return ax

    def plot_doy_by_regions(
        self,
        band: str,
        regions: ee.FeatureCollection,
        label: str = "system:index",
        spatialReducer: str | ee.Reducer = "mean",
        timeReducer: str | ee.Reducer = "mean",
        scale: int = 10000,
        dateProperty: str = "system:time_start",
        colors: list = [],
        ax: Axes | None = None,
    ):
        """Plot the reduced data for each image in the collection by regions for a single band.

        This method is plotting the reduced data for each image in the collection by regions for a single band.

        Parameters:
            band: The band to reduce.
            regions: The regions to reduce the data on.
            label: The property to use as label for each region. Default is "system:index".
            spatialReducer: The name of the reducer or a reducer object to use. Default is "mean".
            timeReducer: The name of the reducer or a reducer object to use. Default is "mean".
            scale: The scale in meters to use for the reduction. default is 10000m
            dateProperty: The property to use as date for each image. Default is "system:time_start".
            colors: The colors to use for the regions. If empty, the default colors are used.
            ax: The matplotlib axes to plot the data on. If None, a new figure is created.

        Returns:
            A matplotlib axes with the reduced values for each region and each day.

        Examples:
            .. code-block:: python

            import ee, geetools

            ee.Initialize()

            collection = (
                ee.ImageCollection("LANDSAT/LC08/C01/T1_TOA")
                .filterBounds(ee.Geometry.Point(-122.262, 37.8719))
                .filterDate("2014-01-01", "2014-12-31")
            )

            regions = ee.FeatureCollection([
                ee.Feature(ee.Geometry.Point(-122.262, 37.8719).buffer(10000), {"name": "region1"}),
                ee.Feature(ee.Geometry.Point(-122.262, 37.8719).buffer(20000), {"name": "region2"})
            ])

            collection.geetools.plot_doy_by_regions("B1", regions, "name", "mean", "mean", 10000, "system:time_start")
        """
        # get the reduced data
        raw_data = self.doyByRegions(
            band, regions, label, spatialReducer, timeReducer, scale, dateProperty
        ).getInfo()

        # transform all the dates strings into int object and reorder the dictionary
        def to_int(d):
            return {int(k): v for k, v in d.items()}

        data = {l: dict(sorted(to_int(raw_data[l]).items())) for l in raw_data}

        # create the plot
        ax = plot_data("doy", data, "Day of Year", colors, ax)

        return ax

    def plot_doy_by_years(
        self,
        band: str,
        region: ee.Geometry,
        reducer: str | ee.Reducer = "mean",
        scale: int = 10000,
        dateProperty: str = "system:time_start",
        colors: list = [],
        ax: Axes | None = None,
    ):
        """Plot the reduced data for each image in the collection by years for a single band.

        This method is plotting the reduced data for each image in the collection by years for a single band.

        Parameters:
            band: The band to reduce.
            region: The region to reduce the data on.
            reducer: The name of the reducer or a reducer object to use. Default is "mean".
            scale: The scale in meters to use for the reduction. default is 10000m
            dateProperty: The property to use as date for each image. Default is "system:time_start".
            colors: The colors to use for the regions. If empty, the default colors are used.
            ax: The matplotlib axes to plot the data on. If None, a new figure is created.

        Returns:
            A matplotlib axes with the reduced values for each year and each day.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                collection = (
                    ee.ImageCollection("LANDSAT/LC08/C01/T1_TOA")
                    .filterBounds(ee.Geometry.Point(-122.262, 37.8719))
                    .filterDate("2014-01-01", "2014-12-31")
                )

                collection.geetools.plot_doy_by_years("B1", ee.Geometry.Point(-122.262, 37.8719).buffer(10000), "mean", 10000, "system:time_start")
        """
        # get the reduced data
        raw_data = self.doyByYears(band, region, reducer, scale, dateProperty).getInfo()

        # transform all the dates strings into int object and reorder the dictionary
        def to_int(d):
            return {int(k): v for k, v in d.items()}

        data = {l: dict(sorted(to_int(raw_data[l]).items())) for l in raw_data}

        # create the plot
        ax = plot_data("doy", data, "Day of Year", colors, ax)

        return ax



"""Toolbox for the `ee.FeatureCollection` class."""
from __future__ import annotations

import ee
import geopandas as gpd
from matplotlib import pyplot as plt
from matplotlib.axes import Axes

from .accessors import register_class_accessor
from .utils import plot_data


@register_class_accessor(ee.FeatureCollection, "geetools")
class FeatureCollectionAccessor:
    """Toolbox for the `ee.FeatureCollection` class."""

    def __init__(self, obj: ee.FeatureCollection):
        """Initialize the FeatureCollection class."""
        self._obj = obj

    def toImage(
        self,
        color: str | ee.String | int | ee.Number = 0,
        width: str | ee.String | int | ee.Number = "",
    ) -> ee.Image:
        """Paint the current FeatureCollection to an Image.

        It's simply a wrapper on Image.paint() method

        Args:
            color: The pixel value to paint into every band of the input image, either as a number which will be used for all features, or the name of a numeric property to take from each feature in the collection.
            width: Line width, either as a number which will be the line width for all geometries, or the name of a numeric property to take from each feature in the collection. If unspecified, the geometries will be filled instead of outlined.
        """
        params = {"color": color}
        width == "" or params.update(width=width)
        return ee.Image().paint(self._obj, **params)

    def addId(
        self, name: str | ee.String = "id", start: int | ee.Number = 1
    ) -> ee.FeatureCollection:
        """Add a unique numeric identifier, starting from parameter ``start``.

        Returns:
            The parsed collection with a new id property

        Example:
            .. code-block:: python

                import ee
                import geetools

                ee.Initialize()

                fc = ee.FeatureCollection('FAO/GAUL/2015/level0')
                fc = fc.geetools.addId()
                print(fc.first().get('id').getInfo())
        """
        start, name = ee.Number(start).toInt(), ee.String(name)

        indexes = ee.List(self._obj.aggregate_array("system:index"))
        ids = ee.List.sequence(start, start.add(self._obj.size()).subtract(1))
        idByIndex = ee.Dictionary.fromLists(indexes, ids)
        return self._obj.map(lambda f: f.set(name, idByIndex.get(f.get("system:index"))))

    def mergeGeometries(self) -> ee.Geometry:
        """Merge the geometries the included features.

        Returns:
            the dissolved geometry

        Example:
            .. code-block:: python

                import ee
                import geetools

                ee.Initialize()

                fc = ee.FeatureCollection("FAO/GAUL/2015/level0")
                fc =fc.filter(ee.Filter.inList("ADM0_CODE", [122, 237, 85]))
                geom = fc.geetools.mergeGeometries()
                print(geom.getInfo())
        """
        first = self._obj.first().geometry()
        union = self._obj.iterate(lambda f, g: f.geometry().union(g), first)
        return ee.Geometry(union).dissolve()

    def toPolygons(self) -> ee.FeatureCollection:
        """Drop any geometry that is not a Polygon or a multipolygon.

        This method is made to avoid errors when performing zonal statistics and/or other surfaces operations. These operations won't work on geometries that are Lines or points. The methods remove these geometry types from GEometryCollections and rremove features that don't have any polygon geometry

        Returns:
            The parsed collection with only polygon/MultiPolygon geometries

        Example:
            .. code-block:: python

                import ee
                import geetools

                ee.Initialize()

                point0 = ee.Geometry.Point([0,0], proj="EPSG:4326")
                point1 = ee.Geometry.Point([0,1], proj="EPSG:4326")
                poly0 = point0.buffer(1, proj="EPSG:4326")
                poly1 = point1.buffer(1, proj="EPSG:4326").bounds(proj="EPSG:4326")
                line = ee.Geometry.LineString([point1, point0], proj="EPSG:4326")
                multiPoly = ee.Geometry.MultiPolygon([poly0, poly1], proj="EPSG:4326")
                geometryCol = ee.Algorithms.GeometryConstructors.MultiGeometry([multiPoly, poly0, poly1, point0, line], crs="EPSG:4326", geodesic=True, maxError=1)

                fc = ee.FeatureCollection([geometryCol])
                fc = fc.geetools.toPolygons()
                print(fc.getInfo())
        """

        def filterGeom(geom):
            geom = ee.Geometry(geom)
            return ee.Algorithms.If(geom.type().compareTo("Polygon"), None, geom)

        def removeNonPoly(feat):
            filteredGeoms = feat.geometry().geometries().map(filterGeom, True)
            proj = feat.geometry().projection()
            return feat.setGeometry(ee.Geometry.MultiPolygon(filteredGeoms, proj))

        return self._obj.map(removeNonPoly)

    def byProperties(
        self,
        featureId: str | ee.String = "system:index",
        properties: list | ee.List = [],
        labels: list = [],
    ) -> ee.Dictionary:
        """Get a dictionary with all feature values for each properties.

        This method is returning a dictionary with all the properties as keys and their values in each feaure as a list.

        .. code-block::

            {
                "property1": {"feature1": value1, "feature2": value2, ...},
                "property2": {"feature1": value1, "feature2": value2, ...},
                ...
            }

        The output remain server side and can be used to create a client side plot.

        Args:
            featureId: The property used to label features. Defaults to "system:index".
            properties: A list of properties to get the values from.
            labels: A list of names to replace properties names. Default to the properties names.

        Returns:
            A dictionary with all the properties as keys and their values in each feaure as a list.

        See Also:
            - :docstring:`ee.FeatureCollection.geetools.byFeatures`
            - :docstring:`ee.FeatureCollection.geetools.plot_by_properties`

        Example:
            .. code-block:: python

                import ee, geetools

                fc = ee.FeatureCollection("FAO/GAUL/2015/level2").limit(10)
                d = fc.geetools.byProperties(["ADM1_CODE", "ADM2_CODE"])
                d.getInfo()
        """
        # get all the id values, they must be string so we are forced to cast them manually
        # the default casting is broken from Python side: https://issuetracker.google.com/issues/329106322
        features = self._obj.aggregate_array(featureId)
        isString = lambda i: ee.Algorithms.ObjectType(i).compareTo("String").eq(0)  # noqa: E731
        features = features.map(lambda i: ee.Algorithms.If(isString(i), i, ee.Number(i).format()))

        # retrieve properties for each feature
        properties = ee.List(properties) if properties else self._obj.first().propertyNames()
        properties = properties.remove(featureId)
        values = properties.map(
            lambda p: ee.Dictionary.fromLists(features, self._obj.aggregate_array(p))
        )

        # get the label to use in the dictionary if requested
        labels = ee.List(labels) if labels else properties

        return ee.Dictionary.fromLists(labels, values)

    def byFeatures(
        self,
        featureId: str | ee.String = "system:index",
        properties: list | ee.List = [],
        labels: list = [],
    ) -> ee.Dictionary:
        """Get a dictionary with all property values for each feature.

        This method is returning a dictionary with all the feature ids as keys and their properties as a dictionary.

        .. code-block::

            {
                "feature1": {"property1": value1, "property2": value2, ...},
                "feature2": {"property1": value1, "property2": value2, ...},
                ...
            }

        The output remain server side and can be used to create a client side plot.

        Args:
            featureId: The property to use as the feature id. Defaults to "system:index". This property needs to be a string property.
            properties: A list of properties to get the values from.
            labels: A list of names to replace properties names. Default to the properties names.

        Returns:
            A dictionary with all the feature ids as keys and their properties as a dictionary.

        See Also:
            - :docstring:`ee.FeatureCollection.geetools.byProperties`
            - :docstring:`ee.FeatureCollection.geetools.plot_by_features`

        Examples:
            .. code-block:: python

                import ee, geetools

                fc = ee.FeatureCollection("FAO/GAUL/2015/level2").limit(10)
                d = fc.geetools.byFeature(featureId="ADM2_CODE", properties=["ADM0_CODE"])
                d.getInfo()
        """
        # compute the properties and their labels
        props = ee.List(properties) if properties else self._obj.first().propertyNames()
        props = props.remove(featureId)
        labels = ee.List(labels) if labels else props

        # create a function to get the properties of a feature
        # we need to map the featureCollection into a list as it's not possible to return something else than a
        # featureCollection mapping a FeatureCollection. very expensive process but we don't have any other choice.
        fc = self._obj.select(propertySelectors=props, newProperties=props)
        fc_list = fc.toList(self._obj.size())
        values = fc_list.map(lambda f: ee.Feature(f).select(props, labels).toDictionary(labels))

        # get all the id values, they must be string so we are forced to cast them manually
        # the default casting is broken from Python side: https://issuetracker.google.com/issues/329106322
        features = self._obj.aggregate_array(featureId)
        isString = lambda i: ee.Algorithms.ObjectType(i).compareTo("String").eq(0)  # noqa: E731
        features = features.map(lambda i: ee.Algorithms.If(isString(i), i, ee.Number(i).format()))

        return ee.Dictionary.fromLists(features, values)

    def plot_by_features(
        self,
        type: str = "bar",
        featureId: str = "system:index",
        properties: list = [],
        labels: list = [],
        colors: list = [],
        ax: Axes | None = None,
        **kwargs,
    ) -> Axes:
        """Plot the values of a ``ee.FeatureCollection`` by feature.

        Each feature property selected in properties will be plotted using the ``featureId`` as the x-axis.
        If no ``properties`` are provided, all properties will be plotted.
        If no ``featureId`` is provided, the "system:index" property will be used.

        Warning:
            This function is a client-side function.

        Args:
            type: The type of plot to use. Defaults to "bar". can be any type of plot from the python lib `matplotlib.pyplot`. If the one you need is missing open an issue!
            featureId: The property to use as the x-axis (name the features). Defaults to "system:index".
            properties: A list of properties to plot. Defaults to all properties.
            labels: A list of labels to use for plotting the properties. If not provided, the default labels will be used. It needs to match the properties length.
            colors: A list of colors to use for plotting the properties. If not provided, the default colors from the matplotlib library will be used.
            ax: The matplotlib axes to use. If not provided, the plot will be send to a new figure.
            kwargs: Additional arguments from the ``pyplot`` function.

        See Also:
            - :docstring:`ee.FeatureCollection.geetools.byFeatures`
            - :docstring:`ee.FeatureCollection.geetools.plot_by_properties`
            - :docstring:`ee.FeatureCollection.geetools.plot_hist`
            - :docstring:`ee.FeatureCollection.geetools.plot`

        Examples:
            .. code-block:: python

                import ee, geetools

                fc = ee.FeatureCollection("FAO/GAUL/2015/level2").limit(10)
                fc.geetools.plot_by_features(properties=["ADM1_CODE", "ADM2_CODE"])
        """
        # Get the features and properties
        props = ee.List(properties) if properties else self._obj.first().propertyNames().getInfo()
        props = props.remove(featureId)

        # get the data from server
        data = self.byProperties(featureId, props, labels).getInfo()

        # reorder the data according to the labels or properties set by the user
        labels = labels if labels else props.getInfo()
        data = {k: data[k] for k in labels}

        return plot_data(type=type, data=data, label_name=featureId, colors=colors, ax=ax, **kwargs)

    def plot_by_properties(
        self,
        type: str = "bar",
        featureId: str = "system:index",
        properties: list | ee.List = [],
        labels: list = [],
        colors: list = [],
        ax: Axes | None = None,
        **kwargs,
    ) -> Axes:
        """Plot the values of a FeatureCollection by property.

        Each features will be represented by a color and each property will be a bar of the bar chart.

        Warning:
            This function is a client-side function.

        Args:
            type: The type of plot to use. Defaults to "bar". can be any type of plot from the python lib `matplotlib.pyplot`. If the one you need is missing open an issue!
            featureId: The property to use as the y-axis (name the features). Defaults to "system:index".
            properties: A list of properties to plot. Defaults to all properties.
            labels: A list of labels to use for plotting the properties. If not provided, the default labels will be used. It needs to match the properties length.
            colors: A list of colors to use for plotting the properties. If not provided, the default colors from the matplotlib library will be used.
            ax: The matplotlib axes to use. If not provided, the plot will be send to a new figure.
            kwargs: Additional arguments from the ``pyplot`` function.

        See Also:
            - :docstring:`ee.FeatureCollection.geetools.byProperties`
            - :docstring:`ee.FeatureCollection.geetools.plot_by_features`
            - :docstring:`ee.FeatureCollection.geetools.plot_hist`
            - :docstring:`ee.FeatureCollection.geetools.plot`

        Examples:
            .. code-block:: python

                import ee, geetools

                fc = ee.FeatureCollection("FAO/GAUL/2015/level2").limit(10)
                fc.geetools.plot_by_properties(xProperties=["ADM1_CODE", "ADM2_CODE"])
        """
        # Get the features and properties
        fc = self._obj
        props = ee.List(properties) if properties else fc.first().propertyNames()
        props = props.remove(featureId)

        # get the data from server
        data = self.byFeatures(featureId, props, labels).getInfo()

        # reorder the data according to the lapbes or properties set by the user
        labels = labels if labels else props.getInfo()
        data = {f: {k: data[f][k] for k in labels} for f in data.keys()}

        return plot_data(type=type, data=data, label_name=featureId, colors=colors, ax=ax, **kwargs)

    def plot_hist(
        self,
        property: str | ee.String,
        label: str = "",
        ax: Axes | None = None,
        color=None,
        **kwargs,
    ) -> Axes:
        """Plot the histogram of a specific property.

        Warning:
            This function is a client-side function.

        Args:
            property: The property to display
            label: The label to use for the property. If not provided, the property name will be used.
            ax: The matplotlib axes to use. If not provided, the plot will be send to the current axes (``plt.gca()``)
            color: The color to use for the plot. If not provided, the default colors from the matplotlib library will be used.
            kwargs: Additional arguments from the ``pyplot.hist`` function.

        See Also:
            - :docstring:`ee.FeatureCollection.geetools.plot_by_features`
            - :docstring:`ee.FeatureCollection.geetools.plot_by_properties`
            - :docstring:`ee.FeatureCollection.geetools.plot`

        Examples:
            .. code-block:: python

                import ee, geetools

                normClim = ee.ImageCollection('OREGONSTATE/PRISM/Norm81m').toBands()
                region = ee.Geometry.Rectangle(-123.41, 40.43, -116.38, 45.14)
                climSamp = normClim.sample(region, 5000)
                climSamp.geetools.plot_hist("07_ppt")
        """
        # gather the data from parameters
        properties, labels = ee.List([property]), ee.List([label])

        # get the data from the server
        data = self.byProperties(properties=properties, labels=labels).getInfo()

        # define the ax if not provided by the user
        if ax is None:
            fig, ax = plt.subplots()

        # gather the data from the data variable
        labels = list(data.keys())
        if len(labels) != 1:
            raise ValueError("Pie chart can only be used with one property")

        kwargs["rwidth"] = kwargs.get("rwidth", 0.9)
        kwargs["color"] = color or plt.get_cmap("tab10").colors[0]
        ax.hist(list(data[labels[0]].values()), **kwargs)
        ax.set_xlabel(labels[0])
        ax.set_ylabel("frequency")

        # customize the layout of the axis
        ax.grid(axis="y")
        ax.set_axisbelow(True)
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)

        # make sure the canvas is only rendered once.
        ax.figure.canvas.draw_idle()

        return ax

    def plot(
        self,
        ax: Axes,
        property: str = "",
        crs: str = "EPSG:4326",
        cmap: str = "viridis",
        boundaries: bool = False,
        color: str = "k",
    ):
        """Plot the featureCollection on a map using the provided property.

        Warning:
            This function is a client-side function.

        Parameters:
            property: The property to use to color the features.
            ax: The axes to plot the map on.
            crs: The CRS to use for the map.
            cmap: The colormap to use for the colors.
            boundaries: Whether to plot the features values or only the boundaries.
            color: The color to use for the boundaries.

        See Also:
            - :docstring:`ee.FeatureCollection.geetools.plot_by_features`
            - :docstring:`ee.FeatureCollection.geetools.plot_by_properties`
            - :docstring:`ee.FeatureCollection.geetools.plot_hist`

        Examples:
            .. code-block:: python

                import ee
                import geetools

                ee.Initialize()

                fc = ee.FeatureCollection("FAO/GAUL/2015/level2").limit(10)
                fig, ax = plt.subplots()
                fc.geetools.plot("ADM2_CODE", ax)
        """
        # get the data from the server
        names = self._obj.first().propertyNames()
        names = names.filter(ee.Filter.stringStartsWith("item", "system:").Not())
        property = property if property != "" else names.get(0).getInfo()
        data = self._obj.select([property]).getInfo()

        # transform the data to a geodataframe and reproject it to the destination crs
        gdf = gpd.GeoDataFrame.from_features(data["features"]).set_crs(4326).to_crs(crs)

        # plot the data on the map either as contours or a valued features
        if boundaries is True:
            gdf.boundary.plot(ax=ax, color=color)
        else:
            gdf.plot(column=property, ax=ax, cmap=cmap)



"""Test the Array class methods."""
import ee


class TestFull:
    """Test the full method."""

    def test_full_with_integers(self):
        full_array = ee.Array.geetools.full(3, 3, 1)
        assert full_array.getInfo() == [[1, 1, 1], [1, 1, 1], [1, 1, 1]]

    def test_full_with_floats(self):
        full_array = ee.Array.geetools.full(3.1, 3.1, 1.0)
        assert full_array.getInfo() == [
            [1.0, 1.0, 1.0],
            [1.0, 1.0, 1.0],
            [1.0, 1.0, 1.0],
        ]

    def test_full_with_ee_numbers(self):
        full_array = ee.Array.geetools.full(ee.Number(3), ee.Number(3), ee.Number(1))
        assert full_array.getInfo() == [[1, 1, 1], [1, 1, 1], [1, 1, 1]]


class TestSet:
    """Test the set method."""

    def test_set_with_integers(self):
        array = ee.Array.geetools.full(3, 3, 1)
        set_array = array.geetools.set(1, 1, 0)
        assert set_array.getInfo() == [[1, 1, 1], [1, 0, 1], [1, 1, 1]]

    def test_set_with_floats(self):
        array = ee.Array.geetools.full(3.1, 3.1, 1.0)
        set_array = array.geetools.set(1, 1, 0.0)
        assert set_array.getInfo() == [
            [1.0, 1.0, 1.0],
            [1.0, 0.0, 1.0],
            [1.0, 1.0, 1.0],
        ]

    def test_set_with_ee_numbers(self):
        array = ee.Array.geetools.full(3, 3, ee.Number(1))
        set_array = array.geetools.set(1, 1, ee.Number(0))
        assert set_array.getInfo() == [[1, 1, 1], [1, 0, 1], [1, 1, 1]]



"""Toolbox for the ``ee.Geometry`` class."""
from __future__ import annotations

import ee

from .accessors import register_class_accessor


@register_class_accessor(ee.Geometry, "geetools")
class GeometryAccessor:
    """Toolbox for the ``ee.Geometry`` class."""

    def __init__(self, obj: ee.Geometry):
        """Initialize the Geometry class."""
        self._obj = obj

    def keepType(self, type: str) -> ee.Geometry:
        """Only keep the geometries of the given type from a GeometryCollection.

        Args:
            type: The type of geometries to keep. Can be one of: Point, LineString, LineRing Polygon.

        Returns:
            .. code-block:: python

                import ee
                import geetools

                ee.Initialize()

                point0 = ee.Geometry.Point([0,0], proj="EPSG:4326")
                point1 = ee.Geometry.Point([0,1], proj="EPSG:4326")
                poly0 = point0.buffer(1, proj="EPSG:4326")
                poly1 = point1.buffer(1, proj="EPSG:4326").bounds(proj="EPSG:4326")
                line = ee.Geometry.LineString([point1, point0], proj="EPSG:4326")
                multiPoly = ee.Geometry.MultiPolygon([poly0, poly1], proj="EPSG:4326")
                geometryCol = ee.Algorithms.GeometryConstructors.MultiGeometry([multiPoly, poly0, poly1, point0, line], crs="EPSG:4326", geodesic=True, maxError=1)
                geom = geometryCol.geetools.keepType('LineString')
                print(geom.getInfo())
        """
        # will raise an error if self is not a GeometryCollection
        error_msg = "This method can only be used with GeometryCollections"
        assert self._obj.type().getInfo() == "GeometryCollection", error_msg

        def filterType(geom):
            geom = ee.Geometry(geom)
            return ee.Algorithms.If(geom.type().compareTo(type), None, geom)

        self._obj
        geometries = self._obj.geometries().map(filterType, True)
        return getattr(ee.Geometry, "Multi" + type)(geometries, self._obj.projection())



"""Test the ``DateRange`` class."""
import ee
import pytest


class TestSplit:
    """Test the ``split`` method."""

    def test_split(self, daterange_instance):
        list = daterange_instance.geetools.split(1, "day")
        first = ee.DateRange(list.get(0)).start()
        last = ee.DateRange(list.get(-1)).end()
        assert list.size().getInfo() == 30
        assert first.format("YYYY-MM-dd").getInfo() == "2020-01-01"
        assert last.format("YYYY-MM-dd").getInfo() == "2020-01-31"

    def split_with_end_outside(self, daterange_instance):
        list = daterange_instance.geetools.split(2, "month")
        first = ee.DateRange(list.get(0)).start()
        last = ee.DateRange(list.get(-1)).end()
        assert list.size().getInfo() == 1
        assert first.format("YYYY-MM-dd").getInfo() == "2020-01-01"
        assert last.format("YYYY-MM-dd").getInfo() == "2020-01-31"


class TestCheckUnit:
    """Test the ``check_unit`` method exception."""

    def test_check_unit(self):
        with pytest.raises(ValueError):
            ee.DateRange.geetools.check_unit("toto")



"""Extra methods for the ``ee.Number`` class."""
from __future__ import annotations

import ee

from .accessors import register_class_accessor


@register_class_accessor(ee.Number, "geetools")
class NumberAccessor:
    """toolbox for the ``ee.Number`` class."""

    def __init__(self, obj: ee.Number):
        """Initialize the Number class."""
        self._obj = obj

    def truncate(self, nbDecimals: int | ee.Number = 2) -> ee.Number:
        """Truncate a number to a given number of decimals.

        Parameters:
            nbDecimals: The number of decimals to truncate to.

        Returns:
            The truncated number.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                n = ee.Number(1.23456).geetools.truncate(3)
                n.getInfo()
        """
        nbDecimals = ee.Number(nbDecimals).toInt()
        factor = ee.Number(10).pow(nbDecimals)
        return self._obj.multiply(factor).toInt().divide(factor)



"""A toolbox to use with Google Earth Engine Python API.

The ``geetools`` package extends the Google Earth Engine Python API with pre-processing and processing tools for the most used satellite platforms by adding utility methods for different Earth Engine Objects that are friendly with the Python method chaining using the geetools namespace.
"""
import ee

# import the accessor namespace
from .accessors import geetools

# it needs to be imported first as it's the mother class
from .ee_computed_object import *

# reproduce older structure of the lib (deprecated)
# will be removed along the deprecation cycle
from . import _deprecated_algorithms as algorithms
from . import _deprecated_composite as composite
from .tools import imagecollection

# then we extend all the other classes
from .ee_asset import Asset
from .ee_date import DateAccessor
from .ee_dictionary import DictionaryAccessor
from .ee_feature import FeatureAccessor
from .ee_feature_collection import FeatureCollectionAccessor
from .ee_filter import FilterAccessor
from .ee_geometry import GeometryAccessor
from .ee_image import ImageAccessor
from .ee_join import JoinAccessor
from .ee_list import ListAccessor
from .ee_number import NumberAccessor
from .ee_string import StringAccessor
from .ee_image_collection import ImageCollectionAccessor
from .ee_initialize import InitializeAccessor
from .ee_authenticate import AuthenticateAccessor
from .ee_array import ArrayAccessor
from .ee_date_range import DateRangeAccessor
from .ee_export import ExportAccessor
from .ee_profiler import Profiler

__title__ = "geetools"
__summary__ = "A set of useful tools to use with Google Earth Engine Python" "API"
__uri__ = "http://geetools.readthedocs.io"
__version__ = "1.8.0"

__author__ = "Rodrigo E. Principe"
__email__ = "fitoprincipe82@gmail.com"

__license__ = "MIT"
__copyright__ = "2017 Rodrigo E. Principe"



"""Test the Date class methods."""
import ee
import pytest

import geetools  # noqa: F401


class TestToDatetime:
    """Test the toDatetime method."""

    def test_to_datetime(self, date_instance):
        datetime = date_instance.geetools.to_datetime()
        assert datetime.year == 2020
        assert datetime.month == 1
        assert datetime.day == 1


class TestGetUnitSinceEpoch:
    """Test the getUnitSinceEpoch method."""

    def test_unit_since_epoch(self, date_instance):
        unit = date_instance.geetools.getUnitSinceEpoch("year")
        assert unit.getInfo() >= 49  # 2020 - 1970

    def test_wrong_unit(self, date_instance):
        with pytest.raises(ValueError):
            date_instance.geetools.getUnitSinceEpoch("foo")


class TestFromEpoch:
    """Test the fromEpoch method."""

    def test_from_epoch(self):
        date = ee.Date.geetools.fromEpoch(49, "year")
        assert date.format("YYYY-MM-DD").getInfo() == "2019-01-01"

    def test_wrong_unit(self):
        with pytest.raises(ValueError):
            ee.Date.geetools.fromEpoch(49, "foo")


class TestFromDOY:
    """Test the fromDOY method."""

    def test_from_doy(self):
        date = ee.Date.geetools.fromDOY(1, 2020)
        assert date.format("YYYY-MM-DD").getInfo() == "2020-01-01"

    def test_wrong_year(self):
        # check GEE can use year < EPOCH
        date = ee.Date.geetools.fromDOY(1, 3)
        assert date.format("YYYY-MM-DD").getInfo() == "0003-01-01"

    def test_wrong_doy(self):
        # check that GEE can use > 365 doy
        date = ee.Date.geetools.fromDOY(367, 2020)
        assert date.format("YYYY-MM-DD").getInfo() == "2021-01-01"


class TestIsLeap:
    """Test the isLeap method."""

    def test_is_leap_1992(self):
        leap = ee.Date("1992-01-01").geetools.isLeap()
        assert leap.getInfo() == 1

    def test_is_leap_2000(self):
        leap = ee.Date("2000-01-01").geetools.isLeap()
        assert leap.getInfo() == 1

    def test_is_leap_1900(self):
        leap = ee.Date("1900-01-01").geetools.isLeap()
        assert leap.getInfo() == 0


class TestNow:
    """Test the now method."""

    def test_now(self):
        date = ee.Date.geetools.now()
        assert date.getInfo() is not None



"""Extra methods for the ``ee.Array`` class."""
from __future__ import annotations

import ee

from .accessors import register_class_accessor


@register_class_accessor(ee.Array, "geetools")
class ArrayAccessor:
    """Toolbox for the ``ee.Array`` class."""

    def __init__(self, obj: ee.Array):
        """Initialize the Array class."""
        self._obj = obj

    # -- alternative constructor -----------------------------------------------
    def full(
        self,
        width: float | int | ee.Number,
        height: float | int | ee.Number,
        value: float | int | ee.Number,
    ) -> ee.Array:
        """Create an array with the given dimensions, initialized to the given value.

        Parameters:
            width: The width of the array.
            height: The height of the array.
            value: The value to initialize the array with.

        Returns:
            An array with the given dimensions, initialized to the given value.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                arr = ee.Array.geetools.full(3, 3, 1)
                arr.getInfo()
        """
        width, height = ee.Number(width).toInt(), ee.Number(height).toInt()
        return ee.Array(ee.List.repeat(ee.List.repeat(value, width), height))

    # -- data maniputlation ----------------------------------------------------
    def set(
        self,
        x: int | ee.Number,
        y: int | ee.Number,
        value: float | int | ee.Number,
    ) -> ee.Array:
        """Set the value of a cell in an array.

        Parameters:
            x: The x coordinate of the cell.
            y: The y coordinate of the cell.
            value: The value to set the cell to.

        Returns:
            The array with the cell set to the given value.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                arr = ee.Array.geetools.full(3, 3, 1)
                arr.geetools.set(1, 1, 0).getInfo()
        """
        xPos, yPos = ee.Number(x).toInt(), ee.Number(y).toInt()
        row = ee.List(self._obj.toList().get(yPos)).set(xPos, ee.Number(value))
        return ee.Array(self._obj.toList().set(yPos, row))



"""Test the Dictionary class methods."""
import ee


class TestFromPairs:
    """Test the fromPairs method."""

    def test_from_pairs_with_list(self):
        d = ee.Dictionary.geetools.fromPairs([["foo", 1], ["bar", 2]])
        assert d.getInfo() == {"foo": 1, "bar": 2}

    def test_from_pairs_with_ee_list(self):
        d = ee.Dictionary.geetools.fromPairs(ee.List([["foo", 1], ["bar", 2]]))
        assert d.getInfo() == {"foo": 1, "bar": 2}


class TestSort:
    """Test the sort method."""

    def test_sort(self):
        d = ee.Dictionary({"foo": 1, "bar": 2}).geetools.sort()
        assert d.getInfo() == {"bar": 2, "foo": 1}


class TestGetMany:
    """Test the getMany method."""

    def test_getMany(self):
        d = ee.Dictionary({"foo": 1, "bar": 2}).geetools.getMany(["foo"])
        assert d.getInfo() == [1]



"""Extra methods for the ``ee.Join`` class."""
from __future__ import annotations

import ee

from .accessors import register_class_accessor


@register_class_accessor(ee.Join, "geetools")
class JoinAccessor:
    """Toolbox for the ``ee.Join`` class."""

    def __init__(self, obj: ee.join):
        """Initialize the Join class."""
        self._obj = obj

    @staticmethod
    def byProperty(
        primary: ee.Collection,
        secondary: ee.Collection,
        field: str | ee.String,
        outer: bool = False,
    ) -> ee.Collection:
        """Join 2 collections by a given property field.

        It assumes ids are unique so uses ee.Join.saveFirst.

        Args:
            primary: the first collection
            secondary: the second collection
            field: the field to join by
            outer: whether to keep non matching features

        Returns:
            the joined collection


        Example:
            .. code-block:: python

                import ee
                import geetools

                ee.Initialize()

                point = ee.Geometry.Point([0,0])
                prop1 = {'id': 1, 'prop_from_fc1': 'I am from fc1'}
                prop2 = {'id': 1, 'prop_from_fc2': 'I am from fc2'}
                fc1 = ee.FeatureCollection([ee.Feature(point, prop1)])
                fc2 = ee.FeatureCollection([ee.Feature(point, prop2)])
                joined = ee.Join.geetools.byProperty(fc1, fc2, 'id')
                joined.getInfo()

        """
        field = ee.String(field)
        primary, secondary = ee.FeatureCollection(primary), ee.FeatureCollection(secondary)
        Filter = ee.Filter.equals(leftField=field, rightField=field)
        join = ee.Join.saveFirst(matchKey="match", outer=outer)
        joined = join.apply(primary, secondary, Filter)

        def cleanJoin(feat):
            primaryProp = feat.propertyNames().remove("match")
            secondaryProp = ee.Feature(feat.get("match")).toDictionary()
            return feat.select(primaryProp).setMulti(secondaryProp)

        return ee.FeatureCollection(joined.map(cleanJoin))



"""Test the Float placeholder object."""
import ee
import pytest


class TestFloat:
    """Test the Float placeholder class."""

    def test_init(self):
        with pytest.raises(NotImplementedError):
            ee.Float()

    def test_name(self):
        assert ee.Float.__name__ == "Float"



"""Toolbox for the ``ee.Authenticate`` function."""
from __future__ import annotations

from contextlib import suppress
from pathlib import Path
from shutil import move
from tempfile import TemporaryDirectory

import ee

from .accessors import register_function_accessor


@register_function_accessor(ee.Authenticate, "geetools")
class AuthenticateAccessor:
    """Create an accessor for the ``ee.Authenticate`` function."""

    @staticmethod
    def new_user(name: str = "", credential_pathname: str = "") -> None:
        """Authenticate the user and save the credentials in a specific folder.

        Equivalent to ee.Authenticate but where the registered user will not be the default one (the one you get when running ee.initialize())

        Args:
            name: The name of the user. If not set, it will reauthenticate default.
            credential_pathname: The path to the folder where the credentials are stored. If not set, it uses the default path

        Example:
            .. code-block:: python

                import ee
                import geetools

                # cannot be displayed in the documentation as the creation
                # of a new user requires user interaction

                # geetools.User.create("secondary")
                # geetools.User.set("secondary")
                # ee.Number(1).getInfo()
        """
        name = f"credentials{name}"
        credential_pathname = credential_pathname or ee.oauth.get_credentials_path()
        credential_path = Path(credential_pathname).parent

        # the authenticate method will write the credentials in the default
        # folder and with the default name. We have to save the existing one in tmp,
        # and then exchange places between the newly created and the existing one
        default = Path(ee.oauth.get_credentials_path())

        with TemporaryDirectory() as dir:
            with suppress(FileNotFoundError):
                move(default, Path(dir) / default.name)
            ee.Authenticate()
            move(default, credential_path / name)
            with suppress(FileNotFoundError):
                move(Path(dir) / default.name, default)

    @staticmethod
    def delete_user(name: str = "", credential_pathname: str = "") -> None:
        """Delete a user credential file.

        Args:
            name: The name of the user. If not set, it will delete the default user
            credential_pathname: The path to the folder where the credentials are stored. If not set, it uses the default path

        Example:
            .. code-block:: python

                import ee
                import geetools

                # cannot be displayed in the documentation as the creation
                # of a new user requires user interaction

                # geetools.User.create("secondary")
                # geetools.User.delete("secondary")
                # geetools.User.set("secondary")
                # will raise an error as the user does not exist anymore
        """
        name = f"credentials{name}"
        credential_pathname = credential_pathname or ee.oauth.get_credentials_path()
        credential_path = Path(credential_pathname).parent
        with suppress(FileNotFoundError):
            (credential_path / name).unlink()

    @staticmethod
    def list_user(credential_pathname: str = "") -> list:
        """return all the available users in the set folder.

        To reach "default" simply omit the ``name`` parameter in the User methods

        Args:
            credential_pathname: The path to the folder where the credentials are stored. If not set, it uses the default path

        Returns:
            A list of strings with the names of the users

        Example:
            .. code-block:: python

                import ee
                import geetools

                geetools.User.list(
        """
        credential_pathname = credential_pathname or ee.oauth.get_credentials_path()
        credential_path = Path(credential_pathname).parent
        files = [f for f in credential_path.glob("credentials*") if f.is_file()]
        return [f.name.replace("credentials", "") or "default" for f in files]

    @staticmethod
    def rename_user(new: str, old: str = "", credential_pathname: str = "") -> None:
        """Rename a user without changing the credentials.

        Args:
            new: The new name of the user
            old: The name of the user to rename
            credential_pathname: The path to the folder where the credentials are stored. If not set, it uses the default path

        Example:
            .. code-block:: python

                import ee
                import geetools

                geetools.user.create("secondary")
                geetools.User.rename("secondary", "new_default")
                geetools.User.list()
        """
        old = f"credentials{old}"
        new = f"credentials{new}"
        credential_pathname = credential_pathname or ee.oauth.get_credentials_path()
        credential_path = Path(credential_pathname).parent
        with suppress(FileNotFoundError):
            (credential_path / old).rename(credential_path / new)



"""A directive to generate an API admonition."""
from __future__ import annotations

from docutils import nodes
from docutils.parsers.rst import directives
from docutils.parsers.rst.directives.admonitions import BaseAdmonition
from sphinx.application import Sphinx
from sphinx.util import logging
from sphinx.util.docutils import SphinxDirective
from sphinx.writers.html5 import HTML5Translator

logger = logging.getLogger(__name__)


class api_node(nodes.Admonition, nodes.Element):
    pass


def visit_api_node(self: HTML5Translator, node: api_node) -> None:
    self.visit_admonition(node)


def depart_api_node(self: HTML5Translator, node: api_node) -> None:
    self.depart_admonition(node)


class APIAdmonitionDirective(BaseAdmonition, SphinxDirective):
    """An API entry, displayed (if configured) in the form of an admonition."""

    node_class = api_node
    has_content = True
    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = False
    option_spec = {
        "class": directives.class_option,
        "name": directives.unchanged,
    }

    def run(self) -> list[nodes.Node]:
        if not self.options.get("class"):
            self.options["class"] = ["admonition-api"]

        (api,) = super().run()
        if isinstance(api, nodes.system_message):
            return [api]
        elif isinstance(api, api_node):
            api.insert(0, nodes.title(text="See API"))
            api["docname"] = self.env.docname
            self.add_name(api)
            self.set_source_info(api)
            self.state.document.note_explicit_target(api)
            return [api]
        else:
            raise RuntimeError  # never reached here


def setup(app: Sphinx) -> dict[str, object]:
    """Add custom configuration to sphinx app.

    Args:
        app: the Sphinx application

    Returns:
        the 2 parallel parameters set to ``True``.
    """
    app.add_directive("api", APIAdmonitionDirective)
    app.add_node(api_node, html=(visit_api_node, depart_api_node))

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""Utils methods for file and asset manipulation in the context of batch processing."""
from __future__ import annotations

import re

import numpy as np
from anyascii import anyascii
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.colors import to_rgba


def format_description(description: str) -> str:
    """Format a name to be accepted as a Task description.

    The rule is:
    The description must contain only the following characters: a..z, A..Z,0..9, ".", ",", ":", ";",
    "_" or "-". The description must be at most 100 characters long.

    Args:
        description: The description to format.

    Returns:
        The formatted description.
    """
    replacements = [
        [[" "], "_"],
        [["/"], "-"],
        [["?", "!", "¿", "*"], "."],
        [["(", ")", "[", "]", "{", "}"], ":"],
    ]

    desc = anyascii(description)
    for chars, rep in replacements:
        pattern = "|".join(re.escape(c) for c in chars)
        desc = re.sub(pattern, rep, desc)  # type: ignore

    return desc[:100]


def format_asset_id(description: str) -> str:
    """Format a name to be accepted as an asset Id.

    The rule is:
    Each segment must contain only the following characters: a..z, A..Z, 0..9, "_" or "-".
    Each segment must be at least 1 character long and at most 100 characters long.

    Args:
        description: The description to format.

    Returns:
        The formatted description.
    """
    replacements = [
        [[" "], "_"],
        [["/"], "-"],
        [["?", "!", "¿", "*"], "."],
        [["(", ")", "[", "]", "{", "}", ";", ":", ",", "."], "_"],
    ]

    desc = anyascii(description)
    for chars, rep in replacements:
        pattern = "|".join(re.escape(c) for c in chars)
        desc = re.sub(pattern, rep, desc)  # type: ignore

    return desc


def plot_data(
    type: str,
    data: dict,
    label_name: str,
    colors: list = [],
    ax: Axes | None = None,
    **kwargs,
) -> Axes:
    """Plotting mechanism used in all the plotting functions.

    It binds the matplotlib capabilities with the data aggregated by different xes.
    the shape of the data should as follows:

    .. code-block::

        {
            "label1": {"properties1": value1, "properties2": value2, ...}
            "label2": {"properties1": value1, "properties2": value2, ...},
            ...
        }

    Args:
        type: The type of plot to use. can be any type of plot from the python lib `matplotlib.pyplot`. If the one you need is missing open an issue!
        data: the data to use as inputs of the graph. please follow the fomrmat specified in the documentation.
        label_name: The name of the property that was used to generate the labels
        property_names: The list of names that was used to name the values. They will be used to order the keys of the data dictionary.
        colors: A list of colors to use for the plot. If not provided, the default colors from the matplotlib library will be used.
        ax: The matplotlib axes to use. If not provided, the plot will be send to a new figure.
        kwargs: Additional arguments from the ``pyplot`` chat type selected.
    """
    # define the ax if not provided by the user
    if ax is None:
        fig, ax = plt.subplots()

    # gather the data from parameters
    labels = list(data.keys())
    props = list(data[labels[0]].keys())
    colors = colors if colors else plt.get_cmap("tab10").colors

    # draw the chart based on the type
    if type == "plot":
        for i, label in enumerate(labels):
            kwargs["color"] = colors[i]
            name = props[0] if len(props) == 1 else "Properties values"
            values = list(data[label].values())
            ax.plot(props, values, label=label, **kwargs)
            ax.set_ylabel(name)
            ax.set_xlabel(f"Features (labeled by {label_name})")
            grid_axis = "y"

    elif type == "scatter":
        for i, label in enumerate(labels):
            kwargs["color"] = colors[i]
            name = props[0] if len(props) == 1 else "Properties values"
            values = list(data[label].values())
            ax.scatter(props, values, label=label, **kwargs)
            ax.set_ylabel(name)
            ax.set_xlabel(f"Features (labeled by {label_name})")
            grid_axis = "y"

    elif type == "fill_between":
        for i, label in enumerate(labels):
            kwargs["facecolor"] = to_rgba(colors[i], 0.2)
            kwargs["edgecolor"] = to_rgba(colors[i], 1)
            name = props[0] if len(props) == 1 else "Properties values"
            values = list(data[label].values())
            ax.fill_between(props, values, label=label, **kwargs)
            ax.set_ylabel(name)
            ax.set_xlabel(f"Features (labeled by {label_name})")
            grid_axis = "y"

    elif type == "bar":
        x = np.arange(len(props))
        width = 1 / (len(labels) + 0.8)
        margin = width / 10
        kwargs["width"] = width - margin
        ax.set_xticks(x + width * len(labels) / 2, props)
        for i, label in enumerate(labels):
            kwargs["color"] = colors[i]
            values = list(data[label].values())
            ax.bar(x + width * i, values, label=label, **kwargs)
        grid_axis = "y"

    elif type == "barh":
        y = np.arange(len(props))
        height = 1 / (len(labels) + 0.8)
        margin = height / 10
        kwargs["height"] = height - margin
        ax.set_yticks(y + height * len(labels) / 2, props)
        for i, label in enumerate(labels):
            kwargs["color"] = colors[i]
            values = list(data[label].values())
            ax.barh(y + height * i, values, label=label, **kwargs)
        grid_axis = "x"

    elif type == "stacked":
        x = np.arange(len(props))
        bottom = np.zeros(len(props))
        ax.set_xticks(x, props)
        for i, label in enumerate(labels):
            kwargs.update(color=colors[i], bottom=bottom)
            values = list(data[label].values())
            ax.bar(x, values, label=label, **kwargs)
            bottom += values
        grid_axis = "y"

    elif type == "pie":
        if len(labels) != 1:
            raise ValueError("Pie chart can only be used with one property")
        kwargs["autopct"] = kwargs.get("autopct", "%1.1f%%")
        kwargs["normalize"] = kwargs.get("normalize", True)
        kwargs["labeldistance"] = kwargs.get("labeldistance", None)
        kwargs["wedgeprops"] = kwargs.get("wedgeprops", {"edgecolor": "w"})
        kwargs["textprops"] = kwargs.get("textprops", {"color": "w"})
        kwargs.update(autopct="%1.1f%%", colors=colors)
        values = [data[labels[0]][p] for p in props]
        ax.pie(values, labels=props, **kwargs)
        grid_axis = "y"

    elif type == "donut":
        if len(labels) != 1:
            raise ValueError("Pie chart can only be used with one property")
        kwargs["autopct"] = kwargs.get("autopct", "%1.1f%%")
        kwargs["normalize"] = kwargs.get("normalize", True)
        kwargs["labeldistance"] = kwargs.get("labeldistance", None)
        kwargs["wedgeprops"] = kwargs.get("wedgeprops", {"width": 0.6, "edgecolor": "w"})
        kwargs["textprops"] = kwargs.get("textprops", {"color": "w"})
        kwargs["pctdistance"] = kwargs.get("pctdistance", 0.7)
        kwargs.update(autopct="%1.1f%%", colors=colors)
        values = [data[labels[0]][p] for p in props]
        ax.pie(values, labels=props, **kwargs)
        grid_axis = "y"

    elif type == "date":
        for i, label in enumerate(labels):
            kwargs["color"] = colors[i]
            x, y = list(data[label].keys()), list(data[label].values())
            ax.plot(x, y, label=label, **kwargs)
        ax.set_xlabel("Date")
        grid_axis = "both"

    elif type == "doy":
        for i, label in enumerate(labels):
            kwargs["color"] = colors[i]
            x, y = list(data[label].keys()), list(data[label].values())
            ax.plot(x, y, label=label, **kwargs)
            ax.set_xlabel("Day of year")
            grid_axis = "both"

    else:
        raise ValueError(f"Type {type} is not (yet?) supported")

    # customize the layout of the axis
    ax.grid(axis=grid_axis)
    ax.set_axisbelow(True)
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.legend(bbox_to_anchor=(1.02, 1), loc="upper left")

    # make sure the canvas is only rendered once.
    ax.figure.canvas.draw_idle()

    return ax



"""Extra methods for the ``ee.Dictionary`` class."""
from __future__ import annotations

import ee

from .accessors import register_class_accessor


@register_class_accessor(ee.Dictionary, "geetools")
class DictionaryAccessor:
    """Toolbox for the ``ee.Dictionary`` class."""

    def __init__(self, obj: ee.Dictionary):
        """Initialize the Dictionary class."""
        self._obj = obj

    # -- alternative constructor -----------------------------------------------
    def fromPairs(self, list: list | ee.List) -> ee.Dictionary:
        """Create a dictionary from a list of [[key, value], ...]] pairs.

        Parameters:
            list: A list of pairs (key, value).

        Returns:
            A dictionary using the pairs.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                d = ee.Dictionary.geetools.fromPairs([["foo", 1], ["bar", 2]])
                d.getInfo()
        """
        list = ee.List(list)
        keys = list.map(lambda pair: ee.List(pair).get(0))
        values = list.map(lambda pair: ee.List(pair).get(1))
        return ee.Dictionary.fromLists(keys, values)

    # -- dictionary operations -------------------------------------------------
    def sort(self) -> ee.Dictionary:
        """Sort the dictionary by keys in ascending order.

        Returns:
            The sorted dictionary.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                d = ee.Dictionary({"foo": 1, "bar": 2}).geetools.sort()
                d.getInfo()
        """
        orderededKeys = self._obj.keys().sort()
        values = orderededKeys.map(lambda key: self._obj.get(key))
        return ee.Dictionary.fromLists(orderededKeys, values)

    def getMany(self, list: list | ee.List) -> ee.List:
        """Extract values from a list of keys.

        Parameters:
            list: A list of keys.

        Returns:
            A list of values.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                d = ee.Dictionary({"foo": 1, "bar": 2, "baz": 3})
                d.geetools.getMany(["foo", "bar"]).getInfo()
        """
        return ee.List(list).map(lambda key: self._obj.get(key))



"""Test the List class methods."""
import ee


class TestProduct:
    """Test the product method."""

    def test_product_with_same_type(self, letter_list, data_regression):
        product_list = letter_list.geetools.product(letter_list)
        data_regression.check(product_list.getInfo())

    def test_product_with_different_type(self, letter_list, int_list, data_regression):
        product_list = letter_list.geetools.product(int_list)
        data_regression.check(product_list.getInfo())


class TestComplement:
    """Test the complement method."""

    def test_complement_with_same_type(self, letter_list, data_regression):
        complement_list = letter_list.geetools.complement(letter_list)
        data_regression.check(complement_list.getInfo())

    def test_complement_with_different_type(self, letter_list, int_list, data_regression):
        complement_list = letter_list.geetools.complement(int_list)
        data_regression.check(complement_list.getInfo())


class TestIntersection:
    """Test the intersection method."""

    def test_intersection_with_same_type(self, letter_list, data_regression):
        intersection_list = letter_list.geetools.intersection(letter_list)
        data_regression.check(intersection_list.getInfo())

    def test_intersection_with_different_type(self, letter_list, int_list, data_regression):
        intersection_list = letter_list.geetools.intersection(int_list)
        data_regression.check(intersection_list.getInfo())


class TestUnion:
    """Test the union method."""

    def test_union_with_duplicate(self, letter_list, data_regression):
        union_list = letter_list.geetools.union(letter_list)
        data_regression.check(union_list.getInfo())

    def test_union_without_dupplicates(self, letter_list, int_list, data_regression):
        union_list = letter_list.geetools.union(int_list)
        data_regression.check(union_list.getInfo())


class TestDelete:
    """Test the delete method."""

    def test_delete(self, letter_list, data_regression):
        deleted_list = letter_list.geetools.delete(1)
        data_regression.check(deleted_list.getInfo())


class TestSequence:
    """Test the sequence method."""

    def test_sequence(self):
        seq = ee.List.geetools.sequence(1, 10)
        assert seq.getInfo() == list(range(1, 11))

    def test_sequence_with_step(self):
        seq = ee.List.geetools.sequence(1, 10, 2)
        assert seq.getInfo() == list(range(1, 11, 2)) + [10]

    def test_sequence_with_uneven_step(self):
        seq = ee.List.geetools.sequence(1, 10, 3)
        assert seq.getInfo() == list(range(1, 10, 3)) + [10]

    def test_sequence_with_0_step(self):
        seq = ee.List.geetools.sequence(1, 10, 0)
        assert seq.getInfo() == list(range(1, 11))


class TestReplaceMany:
    """Test the replaceMany method."""

    def test_replace_many(self, letter_list, data_regression):
        replaced_list = letter_list.geetools.replaceMany({"a": "foo", "c": "bar"})
        data_regression.check(replaced_list.getInfo())


class TestZip:
    """Test the zip method."""

    def test_zip(self, letter_list, data_regression):
        zipped_list = ee.List([letter_list, letter_list]).geetools.zip()
        data_regression.check(zipped_list.getInfo())


class TestToStrings:
    """Test the toStrings method."""

    def test_to_strings(self, mix_list, data_regression):
        strings = mix_list.geetools.toStrings()
        data_regression.check(strings.getInfo())


class TestJoin:
    """Test the join method."""

    def test_join(self, mix_list):
        formatted = mix_list.geetools.join()
        assert formatted.getInfo() == "a, 1, Image"

    def test_join_with_separator(self, mix_list):
        formatted = mix_list.geetools.join(separator="; ")
        assert formatted.getInfo() == "a; 1; Image"



"""Extra methods for the ``ee.Date`` class."""
from __future__ import annotations

from datetime import datetime

import ee

from .accessors import register_class_accessor

EE_EPOCH = datetime(1970, 1, 1, 0, 0, 0)


@register_class_accessor(ee.Date, "geetools")
class DateAccessor:
    """Toolbox for the ``ee.Date`` class."""

    def __init__(self, obj: ee.Date):
        """Initialize the Date class."""
        self._obj = obj

    # -- alternative constructor -----------------------------------------------
    @classmethod
    def fromEpoch(cls, number: int, unit: str = "day") -> ee.Date:
        """Set an the number of units since epoch (1970-01-01).

        Parameters:
            number: The number of units since the epoch.
            unit: The unit to return the number of. One of: ``second``, ``minute``, ``hour``, ``day``, ``month``, ``year``.

        Returns:
            The date as a ``ee.Date`` object.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                d = ee.Date.geetools.fromEpoch(49, 'year')
                d.getInfo()
        """
        cls.check_unit(unit)
        return ee.Date(EE_EPOCH.isoformat()).advance(number, unit)

    @classmethod
    def fromDOY(cls, doy: int, year: int) -> ee.Date:
        """Create a date from a day of year and a year.

        Parameters:
            doy: The day of year.
            year: The year.

        Returns:
            The date as a ``ee.Date`` object.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                d = ee.Date.geetools.fromDOY(1, 2020)
                d.getInfo()
        """
        d, y = ee.Number(doy).toInt(), ee.Number(year).toInt()
        return ee.Date.fromYMD(y, 1, 1).advance(d.subtract(1), "day")

    @classmethod
    def now(cls) -> ee.Date:
        """Create a date on current date.

        Returns:
            The date as a ``ee.Date`` object.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                d = ee.Date.geetools.now()
                d.getInfo()
        """
        return ee.Date(datetime.now().isoformat())

    # -- export date -----------------------------------------------------------
    def to_datetime(self) -> datetime:
        """Convert a ``ee.Date`` to a ``datetime.datetime``.

        Returns:
            The ``datetime.datetime`` representation of the ``ee.Date``.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                d = ee.Date('2020-01-01').geetools.toDatetime()
                d.strftime('%Y-%m-%d')

        """
        return datetime.fromtimestamp(self._obj.millis().getInfo() / 1000.0)

    # -- date operations -------------------------------------------------------
    def getUnitSinceEpoch(self, unit: str = "day") -> ee.Number:
        """Get the number of units since epoch (1970-01-01).

        Parameters:
            unit: The unit to return the number of. One of: ``second``, ``minute``, ``hour``, ``day``, ``month``, ``year``.

        Returns:
            The number of units since the epoch.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                d = ee.Date('2020-01-01').geetools.getUnitSinceEpoch('year')
                d.getInfo()
        """
        self.check_unit(unit)
        return self._obj.difference(EE_EPOCH, unit).toInt()

    def isLeap(self) -> ee.Number:
        """Check if the year of the date is a leap year.

        Returns:
            ``1`` if the year is a leap year, ``0`` otherwise.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                d = ee.Date('2020-01-01').geetools.isLeap()
                d.getInfo()
        """
        year = self._obj.get("year")
        divisibleBy4 = year.mod(4).eq(0)
        divisibleBy100 = year.mod(100).eq(0)
        divisibleBy400 = year.mod(400).eq(0)

        # d400 or (d4 and not d100)
        isLeap = divisibleBy400.Or(divisibleBy4.And(divisibleBy100.Not()))

        return isLeap.toInt()

    # -- helper methods --------------------------------------------------------
    @staticmethod
    def check_unit(unit: str) -> None:
        """Check if the provided value is a valid time unit."""
        if unit not in (units := ["second", "minute", "hour", "day", "month", "year"]):
            raise ValueError(f"unit must be one of: {','.join(units)}")



"""A docstring role to read the docstring from a Python method."""
from __future__ import annotations

import inspect
from functools import reduce

import ee
from docutils import nodes
from sphinx.application import Sphinx
from sphinx.util import logging
from sphinx.util.docutils import SphinxRole

import geetools  # noqa: F401

logger = logging.getLogger(__name__)


class DocstringRole(SphinxRole):
    """The docstring role interpreter."""

    def run(self) -> tuple[list[nodes.Node], list[str]]:
        """Setup the role in the builder context."""
        # retrieve the environment from the node members
        env = self.inliner.document.settings.env
        builder = env.app.builder
        current_doc = self.env.docname

        # extract the members we try to reach from the ee lib
        members = self.text.split(".")[1:]

        # reach the final object using getattr. It will allow us to access the complete information
        #  of the object (docstring, qualname, name, source module)
        try:
            o = reduce(getattr, members, ee)
            modules = inspect.getmodule(o).__name__.split(".")
        except Exception as e:
            logger.warning(f"Failed to retrieve {members}: {e}")
            return [nodes.Text(f"{self.text} not found")], []

        # create the docstring node
        docstring = nodes.Text(f": {o.__doc__.splitlines()[0]}")

        # create a complete link to the object using the url and the name of the object
        target_doc = f"autoapi/{'/'.join(modules)}/{o.__qualname__}"
        refuri = builder.get_relative_uri(current_doc, target_doc)
        inline_node = nodes.literal(self.text, self.text, classes=["py", "py-meth"])
        link = nodes.reference("", "", inline_node, internal=True, refuri=refuri)

        return [link, docstring], []


def setup(app: Sphinx) -> dict[str, object]:
    """Add custom configuration to sphinx application."""
    app.add_role("docstring", DocstringRole())

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""Test all the deprecated methods that have not been kept in the new implementation."""


import pytest

import geetools


class TestImageCollection:
    """Test the deprecated_imagecollection module."""

    def test_linear_function_band(self):
        with pytest.raises(NotImplementedError):
            geetools.imagecollection.linearFunctionBand(None, None, None, None)

    def test_linear_function_property(self):
        with pytest.raises(NotImplementedError):
            geetools.imagecollection.linearFunctionProperty(None, None, None, None)

    def linear_interpolation_property(self):
        with pytest.raises(NotImplementedError):
            geetools.imagecollection.linearInterpolationProperty(None, None, None, None)

    def test_gauss_function_band(self):
        with pytest.raises(NotImplementedError):
            geetools.imagecollection.gaussFunctionBand(None, None, None, None, None)

    def test_gauss_function_property(self):
        with pytest.raises(NotImplementedError):
            geetools.imagecollection.gaussFunctionProperty(None, None, None, None, None)

    def testnormal_distribution_property(self):
        with pytest.raises(NotImplementedError):
            geetools.imagecollection.normalDistributionProperty(None, None, None, None, None)

    def test_normal_distribution_band(self):
        with pytest.raises(NotImplementedError):
            geetools.imagecollection.normalDistributionBand(None, None, None, None, None)

    def test_moving_average(self):
        with pytest.raises(NotImplementedError):
            geetools.imagecollection.moving_average(None)


class TestAlgorithm:
    """Test the deprecated_algorithms module."""

    def test_pansharpenkernel(self):
        with pytest.raises(NotImplementedError):
            geetools.algorithms.pansharpenKernel(None, None)

    def test_pansharpenihsFusion(self):
        with pytest.raises(NotImplementedError):
            geetools.algorithms.pansharpenIhsFusion(None)


class TestComposite:
    """Test the deprecated_composite module."""

    def test_max(self, s2_sr):
        with pytest.deprecated_call():
            geetools.composite.max(s2_sr)

    def test_medoidScore(self, s2_sr):
        with pytest.raises(NotImplementedError):
            geetools.composite.medoidScore(s2_sr)



"""Module holding tools for creating composites."""

import ee
from deprecated.sphinx import deprecated


@deprecated(version="1.5.0", reason="Only used to build the medoid composite.")
def medoidScore(collection, bands=None, discard_zeros=False, bandname="sumdist", normalize=True):
    """Compute a score to reflect 'how far' is from the medoid."""
    raise NotImplementedError(
        "This method was only used to build the medoid composite. "
        "The medoid composite is still available in the lib."
    )


@deprecated(version="1.5.0", reason="Use ee.imageCollection.geetools.medoid instead")
def medoid(collection, bands=None, discard_zeros=False):
    """Medoid Composite. Adapted from https://www.mdpi.com/2072-4292/5/12/6481."""
    return ee.ImageCollection(collection).geetools.medoid()


@deprecated(version="1.5.0", reason="Use ee.ImageCollection.geetools.closestDate instead")
def closestDate(col, clip_to_first=False):
    """Make a composite in which masked pixels are filled with the last available pixel."""
    return ee.ImageCollection(col).geetools.closestDate()


@deprecated(version="1.5.0", reason="Use ee.ImageCollection.geetools.reduceInterval instead")
def compositeRegularIntervals(
    collection,
    interval=1,
    unit="month",
    date_range=(1, 0),
    date_range_unit="month",
    direction="backward",
    start=None,
    end=None,
    composite_function=None,
    composite_args=None,
    composite_kwargs=None,
):
    """Make a composite at regular intervals parsing a composite."""
    return ee.ImageCollection(collection).geetools.reduceInterval(unit=unit)


@deprecated(version="1.5.0", reason="Use ee.ImageCollection.geetools.reduceInterval instead")
def compositeByMonth(
    collection, composite_function=None, composite_args=None, composite_kwargs=None
):
    """Make a composite at regular intervals parsing a composite."""
    return ee.ImageCollection(collection).geetools.reduceInterval(unit="month")


@deprecated(version="1.4.0", reason="Use the vanilla Earth Engine API")
def max(collection, band=None):
    """Make a max composite using the specified band."""
    return collection.max()



"""Extra methods for the ``ee.List`` class."""
from __future__ import annotations

import ee

from .accessors import register_class_accessor


@register_class_accessor(ee.List, "geetools")
class ListAccessor:
    """Toolbox for the ``ee.List`` class."""

    def __init__(self, obj: ee.List):
        """Initialize the List class."""
        self._obj = obj

    def product(self, other: list | ee.List) -> ee.List:
        """Compute the cartesian product of 2 list.

        Values will all be considered as string and will be joined with **no spaces**.

        Parameters:
            other: The list to compute the cartesian product with.

        Returns:
            A list of strings corresponding to the cartesian product.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                l1 = ee.List(["1", "2", "3"])
                l2 = ee.List(["a", "b", "c"])

                l1.geetools.product(l2).getInfo()
        """
        l1 = ee.List(self._obj).map(lambda e: ee.String(e))
        l2 = ee.List(other).map(lambda e: ee.String(e))
        product = l1.map(
            lambda e: l2.map(lambda f: ee.Algorithms.String(e).cat(ee.Algorithms.String(f)))
        )
        return product.flatten()

    def complement(self, other: list | ee.List) -> ee.List:
        """Compute the complement of the current list and the ``other`` list.

        The mthematical complement is the list of elements that are in the current list but not in the ``other`` list and vice-versa.

        Parameters:
            other: The list to compute the complement with.

        Returns:
            A list of strings corresponding to the complement of the current list and the ``other`` list.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                l1 = ee.List(["1", "2", "3"])
                l2 = ee.List(["2", "3", "4"])

                l1.geetools.complement(l2).getInfo()
        """
        l1, l2 = ee.List(self._obj), ee.List(other)
        return l1.removeAll(l2).cat(l2.removeAll(l1))

    def intersection(self, other: list | ee.List) -> ee.List:
        """Compute the intersection of the current list and the ``other`` list.

        The intersection is the list of elements that are in both lists.

        Parameters:
            other: The list to compute the intersection with.

        Returns:
            A list of strings corresponding to the intersection of the current list and the ``other`` list.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                l1 = ee.List(["1", "2", "3"])
                l2 = ee.List(["2", "3", "4"])

                l1.geetools.intersection(l2).getInfo()
        """
        l1, l2 = ee.List(self._obj), ee.List(other)
        return l1.removeAll(l1.removeAll(l2))

    def union(self, other: list | ee.List) -> ee.List:
        """Compute the union of the current list and the ``other`` list.

        This list will drop duplicated items.

        Parameters:
            other: The list to compute the union with.

        Returns:
            A list of strings corresponding to the union of the current list and the ``other`` list.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                l1 = ee.List(["1", "2", "3"])
                l2 = ee.List(["2", "3", "4"])

                l1.geetools.union(l2).getInfo()
        """
        l1, l2 = ee.List(self._obj), ee.List(other)
        return l1.cat(l2).distinct()

    # this method is simply a del but the name is protected in the GEE context
    def delete(self, index: int | ee.Number) -> ee.List:
        """Delete an element from a list.

        Parameters:
            index: The index of the element to delete.

        Returns:
            The list without the element at the given index.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                l = ee.List(["a", "b", "c"])
                l.geetools.delete(1).getInfo()
        """
        index = ee.Number(index).toInt()
        return self._obj.slice(0, index).cat(self._obj.slice(index.add(1)))

    @classmethod
    def sequence(
        cls,
        ini: int | ee.Number,
        end: int | ee.Number,
        step: int | ee.Number = 1,
    ) -> ee.List:
        """Create a sequence from ini to end by step.

        Similar to ``ee.List.sequence``, but if end != last item then adds the end to the end of the resuting list.

        Parameters:
            ini: The initial value of the sequence.
            end: The final value of the sequence.
            step: The step of the sequence.

        Returns:
            A list of numbers corresponding to the sequence.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                l = ee.List.geetools.sequence(0, 10, 2)
                l.getInfo()
        """
        ini, end = ee.Number(ini), ee.Number(end)
        step = ee.Number(step).toInt().max(1)
        return ee.List.sequence(ini, end, step).add(end.toFloat()).distinct()

    def replaceMany(self, replace: dict | ee.Dictionary) -> ee.List:
        """Replace many values in a list.

        Parameters:
            replace: the dictionary with the values to replace. the keys are the values to replace and the values are the new values.

        Returns:
            A list with the values replaced

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                l = ee.List(["a", "b", "c"])
                replace = ee.Dictionary({"a": "foo", "c": "bar"})
                l = l.geetools.replaceMany(replace)
                l.getInfo()
        """
        replace = ee.Dictionary(replace)
        keys = replace.keys()
        list = keys.iterate(lambda k, p: ee.List(p).replace(k, replace.get(k)), self._obj)
        return ee.List(list)  # to avoid returning a COmputedObject

    def join(self, separator: str | ee.String = ", ") -> ee.string:
        """Format a list to a string.

        Same as the join method but elements that cannot be stringified will be returned as the object type.

        Parameters:
            separator: The separator to use.

        Returns:
            A string with the list elements separated by commas.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                l = ee.List(['a', 1, ee.Image(0)])
                l.geetools.join().getInfo()
        """
        return self.toStrings().join(separator)

    def toStrings(self) -> ee.List:
        """Convert elements of a list into Strings.

        If the list contains other elements that are not strings or numbers, it will return the object type. For example, ['a', 1, ee.Image(0)] -> ['a', '1', 'Image'].

        Returns:
            A list of strings corresponding to the elements of the list.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                l = ee.List(["a", 1, ee.Image(0)])
                l.geetools.toStrings().getInfo()
        """
        klasses = ee.List(["Float", "Integer", "String"])

        def getString(el):
            otype = ee.Algorithms.ObjectType(el)
            stringReady = klasses.contains(otype)
            return ee.Algorithms.If(stringReady, ee.Algorithms.String(el), otype)

        return self._obj.map(getString)

    def zip(self) -> ee.List:
        """Zip a list of lists.

        The nested lists need to all have the same size. The size of the first element will be taken as reference.

        Returns:
            A list of lists with the zipped elements

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                l = ee.List([[1,2,3], [4,5,6], [7,8,9]])
                l.geetools.zip().getInfo()
        """
        indices = ee.List.sequence(0, ee.List(self._obj.get(0)).size().subtract(1))
        return indices.map(lambda i: self._obj.map(lambda j: ee.List(j).get(i)))



"""Toolbox for the ``ee.Feature`` class."""
from __future__ import annotations

import ee

from .accessors import register_class_accessor


@register_class_accessor(ee.Feature, "geetools")
class FeatureAccessor:
    """Toolbox for the ``ee.Feature`` class."""

    def __init__(self, obj: ee.Feature):
        """Initialize the class."""
        self._obj = obj

    def toFeatureCollection(self) -> ee.FeatureCollection:
        """Convert a feature composed of a multiGeometry geometry into a FEatureCollection.

        Returns:
            The FeatureCollection

        Example:
            .. code-block:: python

                    import ee
                    import geetools

                    ee.Initialize()

                    geoms = ee.Geometry.MultiPoint([[0,0], [0,1]])
                    feature = ee.Feature(geoms).set("foo", "bar")
                    fc = feature.geetools.toFeatureCollection()
                    print(fc.size().getInfo())
        """
        geoms = self._obj.geometry().geometries()
        fc = geoms.map(lambda g: self._obj.setGeometry(g))
        return ee.FeatureCollection(fc)

    def removeProperties(self, properties: list | ee.List) -> ee.Feature:
        """Remove properties from a feature.

        Args:
            properties : List of properties to remove

        Returns:
            The feature without the properties

        Example:
            .. code-block:: python

                    import ee
                    import geetools

                    ee.Initialize()

                    feature = ee.Feature(None).set("foo", "bar", "baz", "foo")
                    feature = feature.geetools.removeProperties(["foo"])
                    print(feature.getInfo())
        """
        properties = ee.List(properties)
        proxy = ee.Feature(self._obj.geometry())  # drop properties
        return proxy.copyProperties(self._obj, exclude=properties)



"""Test the ``Export`` class."""
import ee
import pytest
from ee.cli.utils import wait_for_task

import geetools  # noqa F401


class TestImageCollection:
    """Test the ``imagecollection`` namespace."""

    @pytest.mark.skip(reason="The export task timeout when to many tests are run at the same time")
    def test_toAsset(self, gee_test_folder):
        task_list = ee.batch.Export.geetools.imagecollection.toAsset(
            imagecollection=self.ic,
            description="ic_to_asset",
            index_property="index",
            assetId=(gee_test_folder / "ic_to_asset").as_posix(),
            region=ee.Geometry.Point([0, 0]).buffer(100).bounds(),
            scale=50,
        )
        [task.start() for task in task_list]
        [wait_for_task(task.id, timeout=50) for task in task_list]

        ic = ee.ImageCollection((gee_test_folder / "ic_to_asset").as_posix())
        assert ic.size().getInfo() == 2

    @property
    def ic(self):
        """Return a test image collection."""
        image_list = [ee.Image(i).set("index", f"image_{i}") for i in range(2)]
        return ee.ImageCollection(image_list)



"""All the process that can be run using nox.

The nox run are build in isolated environment that will be stored in .nox. to force the venv update, remove the .nox/xxx folder.
"""

import nox

nox.options.sessions = ["lint", "test", "docs", "mypy"]


@nox.session(reuse_venv=True)
def lint(session):
    """Apply the pre-commits."""
    session.install("pre-commit")
    session.run("pre-commit", "run", "--all-files", *session.posargs)


@nox.session(reuse_venv=True, name="ci-test")
def ci_test(session):
    """Run all the test using the environment variable of the running machine."""
    session.install(".[test]")
    test_files = session.posargs or ["tests"]
    session.run(
        "pytest",
        "--color=yes",
        "--cov",
        "--cov-report=xml",
        *test_files,
    )


@nox.session(reuse_venv=True)
def test(session):
    """Run all the test using the environment variable of the running machine."""
    session.install(".[test]")
    test_files = session.posargs or ["tests"]
    session.run(
        "pytest",
        "--color=yes",
        "--cov",
        "--cov-report=html",
        *test_files,
    )


@nox.session(reuse_venv=True, name="dead-fixtures")
def dead_fixtures(session):
    """Check for dead fixtures within the tests."""
    session.install(".[test]")
    session.run("pytest", "--dead-fixtures")


@nox.session(reuse_venv=True)
def docs(session):
    """Build the documentation."""
    build = session.posargs.pop() if session.posargs else "html"
    session.install(".[doc]")
    dst, warn = f"docs/_build/{build}", "warnings.txt"
    session.run("sphinx-build", "-v", "-b", build, "docs", dst, "-w", warn)


@nox.session(name="mypy", reuse_venv=True)
def mypy(session):
    """Run a mypy check of the lib."""
    session.install("mypy")
    test_files = session.posargs or ["geetools"]
    session.run("mypy", *test_files)


@nox.session(reuse_venv=True)
def stubgen(session):
    """Generate stub files for the lib but requires human attention before merge."""
    session.install("mypy")
    package = session.posargs or ["geetools"]
    session.run("stubgen", "-p", package[0], "-o", "stubs", "--include-private")
